# coding=utf-8

import math

from .units import UnitDispenser, Unit
from .quantities import Quantity

class SIUnitDispenser(UnitDispenser):
	'''
	A subclass of :class:`UnitDispenser` which prepopulates the unit dispenser
	with SI units and some common other units. For a complete list of supported
	units, please see the "Supported Units" chapter of the python-parameters
	documentation.
	'''

	def init_prefixes(self):
		'''
		This method is called by the :class:`UnitDispenser` constructor, at which
		point this method populates the dispenser object with the SI prefixes. See
		the "Supported Units" chapter of the python-parameters documentation for
		a list of supported prefixes.
		'''
		self._prefixes = [
				("yotta", "Y", 1e24),
				("zepto", "Z", 1e21),
				("exa", "E", 1e18),
				("peta", "P", 1e15),
				("tera", "T", 1e12),
				("giga", "G", 1e9),
				("mega", "M", 1e6),
				("kilo", "k", 1e3),
				("milli", "m", 1e-3),
				("micro", (u"μ", "{mu}"), 1e-6),
				("nano", "n", 1e-9),
				("pico", "p", 1e-12),
				("femto", "f", 1e-15),
				("atto", "a", 1e-18),
				("zepto", "z", 1e-21),
				("yocto", "y", 1e-24)
			]

	def init_units(self):
		'''
		This method is called by the :class:`UnitDispenser` constructor, at which
		point this method populates the dispenser object with the SI units (and some
		other common units). See the "Supported Units" chapter of the
		python-parameters documentation for a list of supported units.
		'''

		# Fundamental SI units
		self \
			+ Unit(["constant", "non-dim", "1"], "", 1.0, prefixable=False) \
			+ Unit(["metre", "meter"], "m", 1.0).set_dimensions(length=1) \
			+ Unit("second", "s", 1.0).set_dimensions(time=1) \
			+ Unit("gram", "g", 1e-3).set_dimensions(mass=1) \
			+ Unit("ampere", "A", 1.0).set_dimensions(current=1) \
			+ Unit("kelvin", "K", 1.0).set_dimensions(temperature=1) \
			+ Unit("mole", "mol", 1.0).set_dimensions(substance=1) \
			+ Unit("candela", "cd", 1.0).set_dimensions(intensity=1) \
			+ Unit("dollar", "$", 1.0, prefixable=False).set_dimensions(currency=1) \
			+ Unit("radian", "rad", 1.0).set_dimensions(angle=1)
		self.basis(mass='kg')

		# Non-linear units
		self \
			+ Unit("decibel", "dB", 1.0)

		self.add_conversion_map("dB", "", lambda v: 10**(v/10.))
		self.add_conversion_map("", "dB", lambda v: 10*math.log(v,10))

		# Angular units
		self \
			+ Unit("degree", [u"°","deg"], 180./math.pi).set_dimensions(angle=1)
		self.add_scaling({'time':-1,'angle':1}, {'time':-1}, 1./2/math.pi)
		self.add_scaling({'angle':1}, {}, 1)

		# Scales
		self \
			+ Unit("mile", "mi", 1609.344).set_dimensions(length=1) \
			+ Unit("yard", "yd", 0.9144).set_dimensions(length=1) \
			+ Unit("foot", "ft", 0.3048, plural="feet").set_dimensions(length=1) \
			+ Unit("inch", "in", 0.0254, plural="inches").set_dimensions(length=1) \
			+ Unit(["centimetre", "centimeter"], "cm", 0.01).set_dimensions(length=1) \
			+ Unit("point", "pt", 2.54e-05).set_dimensions(length=1) \
			+ Unit("angstrom", u"Å", 1e-10).set_dimensions(length=1) \
			+ Unit("astronomical unit", "au", 149597870691.0).set_dimensions(length=1) \
			+ Unit("lightyear", "ly", 9460730472580800.).set_dimensions(length=1)

		# Time
		self \
			+ Unit("year", "year", 31557600.0, prefixable=False).set_dimensions(time=1) \
			+ Unit("month", "month", 2629800.0, prefixable=False).set_dimensions(time=1) \
			+ Unit("fortnight", "fortnight", 1209600.0, prefixable=False).set_dimensions(time=1) \
			+ Unit("week", "week", 604800.0, prefixable=False).set_dimensions(time=1) \
			+ Unit("day", "day", 86400.0, prefixable=False).set_dimensions(time=1) \
			+ Unit("hour", "hour", 3600., prefixable=False).set_dimensions(time=1) \
			+ Unit("minute", "min", 60., prefixable=False).set_dimensions(time=1) \
			+ Unit("hertz", "Hz", 1.).set_dimensions(time=-1)

		# Force
		self \
			+ Unit("newton", "N", 1.).set_dimensions(mass=1, length=1, time=-2)

		# Pressure
		self \
			+ Unit("atm", "atm", 101325.0).set_dimensions(mass=1, length=-1, time=-2) \
			+ Unit("bar", "bar", 100000.0).set_dimensions(mass=1, length=-1, time=-2) \
			+ Unit("pascal", "Pa", 1.).set_dimensions(mass=1, length=-1, time=-2) \
			+ Unit("mmHg", "mmHg", 101325. / 760).set_dimensions(mass=1, length=-1, time=-2) \
			+ Unit("psi", "psi", 6894.757).set_dimensions(mass=1, length=-1, time=-2)

		# Energy
		self \
			+ Unit("joule", "J", 1.).set_dimensions(mass=1, length=2, time=-2) \
			+ Unit("calorie", "cal", 4.1868).set_dimensions(mass=1, length=2, time=-2) \
			+ Unit("electronvolt", "eV", 1.602176487e-19).set_dimensions(mass=1, length=2, time=-2) \
			+ Unit("watt", "W", 1.).set_dimensions(mass=1, length=2, time=-3)

		# Electromagnetism
		self \
			+ Unit("coulomb", "C", 1.0).set_dimensions(current=1, time=1) \
			+ Unit("farad", "F", 1.0).set_dimensions(time=4, current=2, length=-2, mass=-1) \
			+ Unit("henry", "H", 1).set_dimensions(mass=1, length=2, time=-2, current=-2) \
			+ Unit("volt", "V", 1.).set_dimensions(mass=1, length=2, current=-1, time=-3) \
			+ Unit("ohm", u"Ω", 1.).set_dimensions(mass=1, length=2, time=-3, current=-2) \
			+ Unit("siemens", "mho", 1.).set_dimensions(mass=-1, length=-2, time=3, current=2) \
			+ Unit("tesla", 'T', 1.).set_dimensions(mass=1, current=-1, time=-2) \
			+ Unit("gauss", "G", 1e-4).set_dimensions(mass=1, current=-1, time=-2)

		# Volume
		self \
			+ Unit(["litre", "liter"], "L", .001).set_dimensions(length=3) \
			+ Unit("gallon", "gal", 4 * 473176473. / 125000000000).set_dimensions(length=3) \
			+ Unit("quart", "qt", 473176473. / 125000000000).set_dimensions(length=3) \
			+ Unit("weber", "Wb", 1.).set_dimensions(length=2, mass=1, time=-2, current=-1)

		# Temperature
		self \
			+ Unit("fahrenheit", [u"°F","degF"], 9./5).set_dimensions(temperature=1) \
			+ Unit("celsius", [u"°C","degC"], 1.).set_dimensions(temperature=1)

		self.add_conversion_map('fahrenheit','celsius',lambda f: (f - 32)*5./9, absolute=True)
		self.add_conversion_map('fahrenheit','kelvin',lambda f: (f + 459.67)*5./9, absolute=True)
		self.add_conversion_map('fahrenheit','celsius',lambda f: f*5./9, absolute=False)
		self.add_conversion_map('fahrenheit','kelvin',lambda f: f*5./9, absolute=False)

		self.add_conversion_map('celsius','fahrenheit',lambda c: c*9./5 + 32, absolute=True)
		self.add_conversion_map('celsius','kelvin',lambda c: c +  273.15, absolute=True)
		self.add_conversion_map('celsius','fahrenheit',lambda c: c*9./5, absolute=False)
		self.add_conversion_map('celsius','kelvin',lambda c: c, absolute=False)

		self.add_context("cm", hbar=1.05457173e-34)

		self.add_scaling({'mass':1,'length':2,'time':-2}, {'time':-1}, lambda hbar: 1./2/math.pi/hbar, context="cm")

class SIQuantity(Quantity):
	'''
	A subclass of :class:`Quantity` which has a fallback default dispenser of an
	:class:`SIUnitDispenser` rather than an empty :class:`UnitDispenser`. See
	documentation of :class:`Quantity` for more information.
	'''

	def _new(self, value, units, dispenser=None, absolute=False):
		return SIQuantity(value, units, dispenser=self.dispenser if dispenser is None else dispenser, absolute=absolute)

	def _fallback_dispenser(self):
		return SIUnitDispenser()



import math
import errors
import warnings
from functools import total_ordering
import numpy as np

from .units import UnitDispenser, Units
from .text import colour_text
from .utility.compat import UnicodeMixin

@total_ordering
class Quantity(UnicodeMixin):
	'''
	Quantity (value,units=None,absolute=False,dispenser=None)

	A Quantity object represents a physical quantity; that is, one with both
	a value and dimensions. It is able to convert between different united
	representations; and keeps track of units in basic arithmetic.

	:param value: The value of the physical quantity in units of 'units'. Can be any python object conforming to standard numeric operations.
	:type value: Numeric
	:param units: A representation of the units of the object. See documentation of 'Units' for more information.
	:type units: str or Units
	:param absolute: Whether this quantity represents an absolute quantity (a quantity with an absolute reference scale)
		 or a relative one (such as a temperature delta).
	:type absolute: bool
	:param dispenser: The unit dispenser object from which unit objects are drawn. If not specified, a new UnitDispenser object is created.
	:type dispenser: UnitDispenser

	Instantiate a Quantity object:
		To create a new Quantity object simply pass the value, unit representation
		and an instance of UnitDispenser to the Quantity constructor. It is important
		that the units provided are recognised by the UnitDispenser.

		>>> q = Quantity(1, 'ms', dispenser=SIUnitDispenser())

		If no dispenser is provided, it internally defaults to the empty UnitDispenser.
		A subclass named :class:`SIQuantity` is also available which defaults to the
		SIUnitDispenser; a unit dispenser prepopulated with SI units. In the rest
		of this documentation we use :class:`SIQuantity` for brevity.

		.. note:: Instantiating a `UnitDispenser` instance will be the slowest part
			of instantiating a `Quantity`; so in circumstances where performance is required,
			providing an already instantiated `UnitDispenser` instance is a good idea.

	Accessing value and units separately:
		You can extract the value and units of a Quantity object separately by
		accessing the 'value' and 'units' attributes of the object.

		>>> SIQuantity(1,'ms').value
		1
		>>> SIQuantity(1,'ms').units
		ms

	Unit conversion:
		You can convert a Quantity object to another Quantity object with different
		units provided the dimensions agree.

		>>> SIQuantity(1,'m')('km')
		0.001 km
		>>> SIQuantity(1,'g/ns')('kg/s')
		1000000.0 kg/s

		You can also do a unit conversion and keep only the value, using the
		right shift operator:

		>>> SIQuantity(1,'m') >> 'km'
		0.001

	Representing in standard basis:
		As a special case of the unit conversion described above, you can represent
		any Quantity in the basis defined by the UnitDispenser, using the
		:func:`basis` method. The SIUnitDispenser's basis is the fundamental SI
		units.

		>>> SIQuantity(1,'km').basis()
		1000 m
		>>> SIQuantity(1,'J').basis()
		1.0 m^2*kg/s^2

	Arithmetic with Quantity objects:
		Quantity objects support basic arithmetic with the following operations:
		- Addition and subtraction with another Quantity object of the same dimensions (final units taken from first argument)
		- Multiplication and division with another Quantity object, or with a scalar numeric value.
		- Absolute values.
		- Arbitrary powers

		For convenience, Quantity objects will recognise two-tuples as a prototype
		for a Quantity, allowing for shorthand in numeric operations.

		For example:

		>>> SIQuantity(1,'ms') + (2,'s')
		2001.0 ms
		>>> (3,'km') - SIQuantity(2,'m')
		2.998 km
		>>> 2 * SIQuantity(10,'m')
		20 m
		>>> SIQuantity(10,'m') * SIQuantity(2,'s')
		20 s*m
		>>> SIQuantity(10,'m') / SIQuantity(3,'s') # Be careful with integer divison
		3 m/s # Answer is wrong because SIQuantity does not change value types.
		>> abs(SIQuantity(-5,'m'))
		5 m
		>> SIQuantity(5,'m*s')**2
		25 s^2*m^2

	Boolean logic with Quantity objects:
		Quantity objects support the following boolean operators:
		- Testing for equality and inequality
		- Testing relative size using less than and greater than

		As for arithmetic, two-tuples are automatically converted to Quantity objects
		for comparison.

		>>> SIQuantity(2,'m') == (200,'cm')
		True
		>>> SIQuantity(2,'m') != (200,'cm')
		False
		>>> SIQuantity(2,'ns') < SIQuantity(1000,'as')
		False
		>>> SIQuantity(2,'ns') > SIQuantity(1000,'as')
		True
	'''

	def __init__(self, value, units=None, absolute=False, dispenser=None):
		if value is None:
			raise errors.QuantityValueError("A quantity's value must not be None.")
		if isinstance(value, (list, tuple)):
			value = np.array(value)

		self.dispenser = dispenser
		self.value = value
		self.units = units
		self.absolute = absolute

	@property
	def value(self):
		'''
		The value (or magnitude) of the quantity in the current units.

		You can update this value using:

		>>> quantity.value = <value>
		'''
		return self.__value
	@value.setter
	def value(self, value):
		self.__value = value

	@property
	def units(self):
		'''
		The units of the quantity.

		You can update the units (maintaining the current value) using:

		>>> quantity.units = <Units object or string representation>
		'''
		return self.__units
	@units.setter
	def units(self, units):
		if not isinstance(units, Units):
			self.__units = self.dispenser(units)
		else:
			self.__units = units

	@property
	def absolute(self):
		'''
		A boolean specifying whether this unit requires an absolute reference frame.
		For example, many temperature scales do.

		You can update this value using:

		>>> quantity.value = <value>
		'''
		return self.__absolute
	@absolute.setter
	def absolute(self, absolute):
		self.__absolute = absolute

	@property
	def dispenser(self):
		'''
		The `UnitDispenser` instance used to generate units from string representations.

		You can update this reference using:

		>>> quantity.dispenser = <UnitDispenser instance>
		'''
		return self.__dispenser
	@dispenser.setter
	def dispenser(self, dispenser):
		if dispenser is not None and not isinstance(dispenser, UnitDispenser):
			raise ValueError("Quantity objects require a `UnitDispenser` instance.")
		self.__dispenser = dispenser if dispenser is not None else self._fallback_dispenser()

	def basis(self):
		'''
		basis()

		:returns: Quantity object with current value expressed in the basis units of the UnitDispenser.

		For example, for the :class:`SIUnitDispenser`:

		>>> SIQuantity(1,'J').basis()
		1.0 kg*m^2/s^2
		'''
		return self(self.units.basis())

	def _new(self, value, units, dispenser=None, absolute=False):
		return Quantity(value, units, dispenser=self.dispenser if dispenser is None else dispenser, absolute=absolute)

	def _fallback_dispenser(self):
		return UnitDispenser()

	def __call__(self, units, dispenser=None, context=False):
		dispenser = dispenser if dispenser is not None else self.dispenser
		if not isinstance(units, Units):
			units = dispenser(units)
		try:
			return self._new(dispenser.conversion_map(self.units, units, context=context, absolute=self.absolute)(self.value), units, dispenser, absolute=self.absolute)
		except:
			return self._new(self.value * self.units.scale(units, context=context, value=self.value), units, dispenser, absolute=self.absolute)

	def __repr__(self):
		return str(self)

	def __unicode__(self):
		return u"%s %s" % (self.value,  unicode(self.units)) + (u" (abs)" if self.absolute else u"")

	# Arithmetic
	def __add__(self, other, reverse=False):
		if other == 0:
			return self._new(self.value, self.units)
		elif type(other) is tuple and len(other) == 2:
			other = self._new(*other)
		elif isinstance(other, Units):
			raise ValueError("Invalid operation")
		elif not isinstance(other, Quantity):
			other = self._new(other, None)

		abs = self.absolute and not other.absolute or not self.absolute and other.absolute
		if reverse:
			scale = self.units.scale(other.units)
			return self._new(scale * self.value + other.value, other.units, absolute=abs)
		else:
			scale = other.units.scale(self.units)
			return self._new(self.value + scale * other.value, self.units, absolute=abs)

	def __radd__(self, other):
		return self.__add__(other, reverse=True)

	def __sub__(self, other, reverse=False):
		if other == 0:
			return self._new(self.value, self.units)
		elif type(other) is tuple and len(other) == 2:
			other = self._new(*other)
		elif isinstance(other, Units):
			raise ValueError("Invalid operation")
		elif not isinstance(other, Quantity):
			other = self._new(other, None)

		abs = self.absolute and not other.absolute or not self.absolute and other.absolute
		if reverse:
			scale = self.units.scale(other.units)
			return self._new(-scale * self.value + other.value, other.units, absolute=abs)
		else:
			scale = other.units.scale(self.units)
			return self._new(self.value - scale * other.value, self.units, absolute=abs)

	def __rsub__(self, other):
		return self.__sub__(other, reverse=True)

	def __abs__(self):
		return self._new(abs(self.value), self.units)

	def __mul__(self, other):
		if type(other) is tuple and len(other) == 2:
			other = self._new(*other)
		elif isinstance(other, Units):
			other = 1.0 * other
		try:
			abs = self.absolute and (self.units.dimensions == {} or other.units.dimensions == {})
			units = self.units * other.units
			return self._new(self.value * other.value, units, absolute=abs)
		except AttributeError:
			return self._new(self.value * other, self.units, absolute=self.absolute)

	def __rmul__(self, other):
		return self.__mul__(other)

	def __div__(self, other):
		if type(other) is tuple and len(other) == 2:
			other = self._new(*other)
		elif isinstance(other, Units):
			other = 1.0 * other

		if isinstance(other, Quantity):
			if self.absolute or other.absolute:
				raise ValueError("Cannot divide absolute quantities.")
			absolute = self.absolute and (self.units.dimensions == {} or other.units.dimensions == {})
			units = self.units / other.units
			return self._new(self.value / other.value, units, absolute=absolute)

		return self._new(self.value / other, self.units, absolute=self.absolute)

	def __truediv__(self, other):
		return self.__div__(other)

	def __rdiv__(self, other):
		if type(other) is tuple and len(other) == 2:
			other = self._new(*other)
			return other / self
		elif isinstance(other, Units):
			other = 1.0 * other
			return other / self
		return self._new(other / self.value, 1 / self.units, absolute=self.absolute)

	def __rtruediv__(self, other):
		return self.__rdiv__(other)

	def __pow__(self, other):
		if isinstance(other, Quantity):
			other = other("").value
		return self._new(self.value ** other, self.units ** other)

	# Duplicate functionality (as in __cmp__) to allow for comparison with non Quantity objects
	def __eq__(self,other):
		if type(other) is tuple and len(other) == 2:
			other = self._new(*other)
		if isinstance(other, Quantity):
			scale = self.units.scale(other.units)
			if self.__truncate(self.value) == self.__truncate(other.value / scale) and self.absolute==other.absolute:
				return True
		return False

	def __ne__(self,other):
		return not self.__eq__(other)
	
	def __lt__(self,other): # Python 3 and newer ignore __cmp__
		return self.__cmp__(other) == -1

	def __cmp__(self, other):
		if type(other) is tuple and len(other) == 2:
			other = self._new(*other)
		if isinstance(other, Quantity):
			scale = self.units.scale(other.units)
			if self.__truncate(self.value) == self.__truncate(other.value / scale):
				return 0
			elif self.__truncate(self.value) > self.__truncate(other.value / scale):
				return 1
			return -1
		raise ValueError("Unknown comparison between Quantity and object of type %s." % (type(other)))

	def __truncate(self, value):
		if value == 0:
			return value
		return round(value, int(-math.floor(math.log(abs(value), 10)) + 10))

	def __rshift__(self, str_units):
		return self(str_units).value

	# numpy compatibility

	__numpy_units_in = {
			# Trigonometry
			'sin': 'rad', 'cos': 'rad', 'tan': 'rad',
			'sinh': 'rad', 'cosh': 'rad', 'tanh': 'rad',
			'arcsin': '', 'arccos': '', 'arctan': '',
			'arcsinh': '', 'arccosh': '', 'arctanh': '',
			# Common functions
			'exp': '', 'expm1': '', 'expm2': '',
			'log': '', 'log10': '', 'log1p': '', 'log2': '',
			# Others
			'radians': 'deg', 'degrees': 'rad',
			'deg2rad': 'deg', 'rad2deg': 'rad',
			'logaddexp': '', 'logaddexp2': ''
		}

	__numpy_units_out = {
			# Trigonometry
			'sin': '', 'cos': '', 'tan': '',
			'sinh': '', 'cosh': '', 'tanh': '',
			'arcsin': 'rad', 'arccos': 'rad', 'arctan': 'rad',
			'arcsinh': 'rad', 'arccosh': 'rad', 'arctanh': 'rad',
			'arctan2': 'rad',
			# Others
			'radians': 'rad', 'degrees': 'deg',
			'deg2rad': 'rad', 'rad2deg': 'deg'
		}

	__numpy_units_power = {
			'sqrt': 0.5, 'square': 2, 'reciprocal': -1
		}

	__numpy_units_whatever = [
			'remainder'
	]

	__numpy_ufuncs = tuple(__numpy_units_in.keys()) + tuple(__numpy_units_out.keys()) +\
						tuple(__numpy_units_power.keys()) + tuple(__numpy_units_whatever)

	__array_priority__ = 1000

	def __getattr__(self, attr):
		if attr.startswith('__array_'):
			return getattr(np.asarray(self.value), attr)
		return object.__getattribute__(self, attr)

	def __array_prepare__(self, array, context=None):
		
		ufunc, objs, domain = context

		if ufunc.__name__ in self.__numpy_ufuncs and domain == 0:
			# Cannot deal with more than one ufunc at a time
			try:
				if self.__handling:
					raise Exception('Cannot handle nested ufuncs.')
			except:
				pass

			self.__handling = context

		return array

	def __array_wrap__(self, array, context=None):

		try:
			ufunc, objs, domain = context

			if ufunc.__name__ not in self.__numpy_ufuncs:
				warnings.warn("ufunc '%s' not explicitly understood. Attempting to apply anyway." % ufunc.__name__)

			if ufunc.__name__ in self.__numpy_units_in:
				objs = list(map(lambda x: x(self.__numpy_units_in[ufunc.__name__] if isinstance(x, Quantity) else x), objs))

			rv = ufunc(*[v.value if isinstance(v,Quantity) else v for v in objs])

			out_units = self.__dispenser(self.__numpy_units_out.get(ufunc.__name__, objs[0].units))
			out_units **= self.__numpy_units_power.get(ufunc.__name__, 1)

			r =  self._new(rv, self.__numpy_units_out.get(ufunc.__name__, out_units))

			return r
		except Exception as e:
			raise e
		finally:
			self.__handling = None

	def __long__(self):
		return long(self("").value)

	def __int__(self):
		return int(self("").value)

	def __float__(self):
		return float(self("").value)

	def __complex__(self):
		return complex(self("").value)



from __future__ import print_function

from . import errors
from . import physical_constants
from .definitions import SIUnitDispenser
from .iteration import RangesIterator
from .quantities import Quantity
from .text import colour_text
from .units import Units, Unit
from .utility.compat import str_types

import copy
import imp
import inspect
import numpy as np
import re
import sympy
import sympy.abc
import types
import warnings


class Parameters(object):
	"""
	Parameters(dispenser=None, default_scaled=True, constants=False)

	:class:`Parameters` is the main class in the :mod:`parameters` package, and
	acts to organise and manage potentially interdependent physical quantities.
	The functionality of this class includes:

	- Non-dimensionalisation of parameters
	- Managing parameters which are functions of other parameters
	- Performing unit conversions of parameters
	- Iterating over ranges of parameters
	- Putting bounds on parameters

	A lot of the functionality of the :class:`Parameters` class is handled in
	"magic" methods, meaning that you will not necessarily find what you want to
	do by looking at the method documentation alone. In the rest of the class
	documentation, we cover the use of Parameters magic methods, alluding to
	the public methods where appropriate.

	:param dispenser: Should provide None or a custom unit dispenser. If None is
		provided, Parameters will instantiate an SIUnitDispenser; which hosts
		the standard units relative to an SI basis.
	:type dispenser: UnitDispenser or None
	:param default_scaled: :python:`True` when Parameters should by default return scaled non-dimensional parameters. :python:`False` otherwise.
	:type default_scaled: bool
	:param constants: :python:`True` when Parameters should import the physical constants when the internal :class:`UnitDispenser` is of type :class:`SIUnitDispenser`. :python:`False` otherwise.
	:type constants: bool

	Initialising a Parameters Instance:
		Initialising a :class:`Parameters` instance with the default configuration is simple:

		>>> p = Parameters()

		If you want to use your own :class:`UnitDispenser`, use united parameters
		by default, and preload the :class:`Parameters` instance with a set of
		physical constants, you could use:

		>>> u = SIUnitDispenser()
		>>> p = Parameters(dispenser=u, default_scaled=False, constants=True)

	Seeing what is stored in a Parameters instance:
		To see an overview of the parameter object; simply use:

		>>> p.show() # See show documentation for more.

		To get a list of parameters stored, use:

		>>> list(p)

		To check if a parameter is stored:

		>>> is_stored = 'x' in p # Will be True if 'x' is in p

	Parameter Definition:
		There are several ways to define parameters, each of which is shown below:

		>>> p(x=1)
		>>> p << {'x':1}
		>>> p['x'] = 1
		>>> p.x = 1

		The first two methods generalise to defining multiple parameters at once,
		whereas the second two do not. The first and third set parameters *after*
		first interpreting the value given (in a manner to be described below),
		whereas the others do not.

		.. note:: The last method will even when the parameter name clashes with a method of :class:`Parameters`, which you will be warned about when you define such a parameter. It will not be possible, however, to retrieve such a parameter using attribute notation later. Fortunately, "interpretation" makes no difference when retrieving parameters, and so the other methods will work as expected.

		Let us begin with the properties that all of these methods share.

		* Parameter names must start with a letter; and may contain any number of additional letters, numbers, and underscores.
		* Parameter values may be:
			- any numeric type (include :class:`complex`).
			- a string that represents a symbolic expression.
			- a function with zero or more arguments, each of which being the name of a parameter.
			- a two-tuple of any of the above types with a valid unit (see the Supported Units section of the Parameters documentation for supported SI units in a :class:`SIUnitDispenser`).
			- a :class:`Quantity` instance

		Here are some examples of valid parameter-value combinations:

		>>> p.x=1.42
		>>> p.alpha = 'sin(x)'
		>>> p.beta2 = lambda x, alpha: math.sqrt(x+alpha)
		>>> p.y = (32.1, '{mu}eV')
		>>> p.q_z_ = ('x+2*beta2', 'ms')

		When a unit is not specified, the parameter is assumed to be a
		scaled version of the parameter. If the unit has never been set,
		then a 'constant' unit is assumed; otherwise the older unit is assumed.

		You can specify the units of a parameter, without specifying its
		value:

		>>> p & {'x':'ms'}

		.. note:: If you do this when the parameter has already been set, the parameter value will be adjusted to maintain a constant physical value. If a parameter has been set, and you try to change the units to another which is dimensionally incompatible (e.g. 'm' -> 'ms'), this will result in a :python:`UnitConversionError` being raised. You should first ask the :class:`Parameters` instance to "forget" this value, as described shortly.

		Defining multiple parameters at once is easy with the first two methods:

		>>> p(x=1,y=(2,'s'),z=lambda x,y: x+y)
		>>> p << {'x':1,'y':(2,'s'),'z': lambda x,y: x+y }

		Note, though, that these two statements are **not** equivalent, as described
		in the next section.

	Interpreted vs Non-interpreted Parameter Definition:
		In the previous section, a distinction was made between methods of
		defining parameters that first "interpret" the value and those that do
		not. In particular: :python:`p(?=...)` and :python:`p[?] = ...` were said
		to first interpret the value, whereas :python:`p << {?: ...}` and
		:python:`p.? = ...` did not.

		The difference amounts to whether or not functions are first evaluated
		before being saved as a parameter value. For example:

		>>> p['x'] = lambda y: y**2 # Evaluated immediately
		>>> p.is_function('y')
		False
		>>> p.x = lambda y: y**2 # Evaluated only when x is determined
		>>> p.is_function('x')
		True

		Succinctly, interpreted parameter definitions set the parameter to the value
		of the function at time of definition; whereas non-interpreted parameter
		definitions cause the parameter to be evaluated in whatever context it finds
		itself in the future.

	Parameter Extraction:
		Just as there are several ways to define parameters, there are several
		ways to extract them. They are:

		>>> p('x')
		>>> p.x
		>>> p['x']

		Each of the above returns either a single non-dimensional value or a
		:class:`Quantity` depending upon whether default_scaled was passed as True
		or False respectively in the parameter instance. The first of these allows for multiple
		parameters to be extracted, whereas the second two do not. Additionally,
		the first of these also allows for temporarily overriding the parameter
		context. For example:

		>>> p('x','y',z=(10,'ms'))

		The above example returns a dictionary which contains a mapping from 'x'
		and 'y' to their respective values; evaluated in the context of 'z' being
		overridden temporarily with a value of 10 ms. At the end of this evaluation,
		'z' maintains whatever value it had before. This only affects the value
		of requested variables if it is a function of those variables.

		.. note:: As in most functions in this class that accept multiple parameter arguments, a dictionary is returned if there are two or more parameters; otherwise the value of the single requested parameter is returned. You can force a dictionary response by encapsulating this parameter in a list. For example::

			>>> p(['x'])

		As mentioned, if the :class:`Parameter` instance was instantiated with
		:python:`default_scaled` set to :python:`True`, then all of the parameters
		requested by name will return a non-dimensionalised number; or if it was
		:python:`False`, a :class:`Quantity` object. To invert this default, simply
		add a '_' to the variable name. For example:

		>>> p('_x')
		>>> p._x
		>>> p['_x']

		As a shorthand for extracting multiple parameters, you can use the following
		special method:

		>>> p._('x','y',z=(10,'ms'))

		This is identical in function to:

		>>> p('_x','_y',z=(10,'ms'))

		Although redundant, this special method can also set parameter values just
		as calling the :class:`Parameter` instance does:

		>>> p._(x=10)

	Parameter Interdependencies:
		As already discussed, parameters within a :class:`Parameters` instance
		can depend on values of other parameters by setting their value to be a
		function or symbolic expression. In this section we describe how to best
		take advantage of this functionality.

		Firstly, the function may either be a full function, method or lambda expression.

		Secondly, the function must take as arguments only parameter names which
		exist in the :class:`Parameters` instance at the time that it is executed.
		Thus, "non-interpreted" parameter definitions can be set to include variables
		that are only defined at runtime; such as the current time in a
		numerical integration. Beyond that, there are no restrictions, except that
		the function must return a value that is understood (this can both a numeric
		quantity, or a united quantity tuple, or a :class:`Quantity` object).

		Thirdly, by using a variable name prepended with an underscore in the
		function declaration, you can access both non-dimensional and dimensional
		quantities in your function.

		Fourthly, you can cause your function to be invertible by adding a reference
		to the variable name you are defining to the end of the list of function
		arguments, with a default value. When that variable is specified, you should
		return an ordered list/tuple of updated parameters for the other parameters in the
		function declaration. If there is only one other, it is sufficient to
		simply return it. For example:

		>>> p.x = lambda y,x=None: y**2 if x is None else x**0.5

		>>> p << {'y': 1, 'z': 1, 'x': lambda y,z,x=None: y+z if x is None else [y,x-y]}
		>>> p('x')
		2.0
		>>> p('z',x=10)
		9.0

		Lastly, parameter relationships can chained and as deep as you like.
		For example, x could depend on y, which could depend on z, and so on.
		The :class:`Parameters` will ensure that there are no dependency loops.

	Removing Parameters:
		To remove a parameter, simply use the forget method:

		>>> p.forget('x','y',...)

		As many parameters as you like can be specified in one function call.

	Parameter Units and Scaling:
		Custom units can be added by using the :func:`unit_add` method, and
		scaling used in the non-dimensionalisation process can be extracted
		for a given unit using the :func:`unit_scaling` method.

		You can customise the scaling used in the non-dimensionalisation process
		by changing the scaling of the various different dimensions using
		the :func:`scaling` method. By default, units are scaled relative
		to the SI fundamental units. The current scaling for a dimension
		can also be extracted using this function.

	Unit Conversion:
		Please see the documentation for the :func:`convert` method.

	Parameter Bounds:
		Please see the documentation for the :func:`bounds` and :func:`set_bounds`
		methods.

	Parameter Ranges:
		It is often useful to iterate over a range of parameter values. You can
		generate such ranges using the :func:`range` method. You can also have
		the iteration process handled for you, as documented in the
		:func:`range_iterator` method.

	Physical Constants:
		If :python:`constants` was :python:`True` when the :class:`Parameters`
		instance was initialised, and a :class:`SIUnitDispenser` is being used
		(as it is by default), then the parameter list is prepopulated with a list
		of physical constants. Please see the "Physical Constants" section of the
		*python-parameters* documentation.

	Loading and Saving Parameter Sets:
		To load a parameter set into a Parameters instance, use the classmethod
		:func:`load`. See its documentation for more information.

		To save your existing parameters, use:

		p >> "filename.py"

		Note that parameters that are dependent on other parameters will not survive
		this transition, and will be saved as static values.

	Parameter Contexts:
		Parameters objects support Python's "with" syntax. Upon exiting a "with"
		environment, any changes made will be reset to before re-entering the
		parent environment.

		>>> with p:
		>>> 	p(x=1)
		>>> p('x') # Returns value of x before entering the with environment.
	"""

	def __init__(self, dispenser=None, default_scaled=True, constants=False):
		self.__parameters_spec = {}
		self.__parameters = {}
		self.__parameters_bounds = None
		self.__scalings = {}
		self.__units = dispenser if dispenser is not None else SIUnitDispenser()
		self.__units_custom = []
		self.__default_scaled = default_scaled

		self.__cache_deps = {}
		self.__cache_sups = {}
		self.__cache_scaled = {}
		self.__cache_funcs = {}

		self.__scaling_cache = {}

		if constants and isinstance(self.__units, SIUnitDispenser):
			self(**physical_constants.constants)

	############## PARAMETER OBJECT CONFIGURATION ##############################
	def __add__(self, other):
		self.unit_add(other)
		return self

	def unit_add(self, *args, **kwargs):
		'''
		unit_add(*args, **kwargs)

		:param args: A length 1 sequence of Unit objects, or args to pass to the :class:`Unit` constructor.
		:type args: tuple of mixed type
		:param kwargs: Keyword arguments to pass to the :class:`Unit` constructor.
		:type kwargs: dict of mixed type

		This method allows you to add custom units to the :class:`UnitDispenser` object
		which generates :class:`Units` objects on demand. These additional units
		can override existing units, or add entirely new units. The :python:`args`
		and :python:`kwargs` passed to this function are essentially passed directly
		to the :class:`Unit` constructor:

		>>> Unit(*args, **kwargs)

		Unless:
		- :python:`args` contains a single :class:`Unit` instance, it is directly added the the :class:`UnitDispenser` used by :class:`Parameters`.
		- :python:`args` contains a single :class:`dict` instance, in which case the unit :python:`Unit(**args[0])` is add to the :class:`UnitDispenser`.

		For more information regarding the parameters which can passed to the :class:`Unit`
		constructor, see documentation for :class:`Unit`.
		'''
		if len(args) == 1:
			if isinstance(args[0], Unit):
				unit = args[0]
			elif isinstance(args[0], dict):
				unit = Unit(**args[0])
			else:
				raise ValueError("Invalid unit type to add: %s" % args[0])
		else:
			unit = Unit(*args, **kwargs)
		self.__units.add(unit)
		self.__units_custom.append(unit)

	def set_units_context(self, *name, **params):
		self.__units.set_context(*name,**params)

	@property
	def units_context(self):
		return self.__units.context

	def scaling(self, *args, **kwargs):
		'''
		scaling(*args, **kwargs)

		:param args: A (possibly empty) sequence of dimensions to query.
		:type args: tuple of str
		:param kwargs: A (possibly empty) specification of scales for various dimensions.
		:type kwargs: dict of str/Units

		:returns: The scaling associated with dimensions listed in :python:`args`.

		It is usually the case in numerical applications that one is only interested
		in dealing with non-dimensionalised quantities, rather than quantities
		with units. By default, Parameters instances will return non-dimensional
		quantities unless negated by the underscore (as described elsewhere).
		This method sets the reference scales which determine the
		non-dimensionalisation. By default, the SI fundamental units are used,
		as defined in :class:`SIUnitDispenser`.

		For example:

		>>> p.scaling(length='m') # This will set the base length unit to 'm'
		>>> p( (1,'m') )
		1.
		>>> p.scaling(time='ns') # This will set base time unit to 'ns'
		>>> p( (1,'s') )
		1e9

		Obviously, this does not affect units returned in united form.

		>>> p._( (1,'s') )
		1 s
		'''
		self.__scaling_cache = {}

		for arg in kwargs:
			if arg in self.__units.dimensions:
				scale = self.__get_quantity(kwargs[arg], param=arg)
				if scale.units.dimensions == {arg: 1}:
					self.__scalings[arg] = scale
				else:
					raise errors.ScalingUnitInvalidError("Dimension of scaling (%s) is wrong for %s." % (scale.units, arg))
			else:
				raise errors.ScalingDimensionInvalidError("Invalid scaling dimension %s." % arg)

		if len(args) == 1:
			return self.__scalings.get(args[0], Quantity(1, self.__units.basis()[args[0]], dispenser=self.__units))
		if len(args) > 0:
			output = {}
			for arg in args:
				output[arg] = self.__scalings.get(arg, Quantity(1, self.__units.basis()[arg], dispenser=self.__units))
			return output

	def unit_scaling(self, *params):
		'''
		unit_scaling(*params)

		:param params: A sequence of params for which to query internal non-dimensionalisation scaling.
		:type params: tuple

		:returns: A dictionary of unit scalings, or, if :python:`params` is of length 1, a single numeric scaling factor.

		This method returns the internal non-dimensionalisation factor used to
		scale united values. For example, using the default :class:`SIUnitDispenser`:

		>>> p.unit_scaling('m')
		1.0
		>>> p.unit_scaling('km')
		0.001
		>>> p.unit_scaling('eV')
		6.241509647120417e+18

		The results should be interpreted as: non-dim_value = united_value / unit_scaling .
		'''
		r = {}
		for param in params:
			r[param] = self.__unit_scaling(param)
		if len(params) == 1:
			return r[params[0]]
		return r

	def __get_unit(self, unit):

		if isinstance(unit, str_types):
			return self.__units(unit)

		elif isinstance(unit, Units):
			return unit

		raise errors.UnitInvalidError("No coercion for %s to Units." % unit)

	################## ENABLE USE WITH 'with' ####################################

	def __enter__(self):
		try:
			self.__context_save
		except:
			self.__context_save = []

		self.__context_save.append({
			'parameters_spec': copy.copy(self.__parameters_spec),
			'parameters': copy.copy(self.__parameters),
			'parameters_bounds': copy.copy(self.__parameters_bounds),
			'scalings': copy.copy(self.__scalings),
			'units': copy.copy(self.__units),
			'units_custom': copy.copy(self.__units_custom),
			'default_scaled': copy.copy(self.__default_scaled),
		})

	def __exit__(self, type, value, traceback):

		context = self.__context_save.pop()

		# Restore context
		self.__parameters_spec = context['parameters_spec']
		self.__parameters = context['parameters']
		self.__parameters_bounds = context['parameters_bounds']
		self.__scalings = context['scalings']
		self.__units = context['units']
		self.__units_custom = context['units_custom']
		self.__default_scaled = context['default_scaled']

		# Remove context
		if len(self.__context_save) == 0:
			del self.__context_save

		# Clear cache
		self.__cache_deps = {}
		self.__cache_sups = {}
		self.__cache_scaled = {}
		self.__scaling_cache = {}

	############# PARAMETER RESOLUTION #########################################
	def __get_pam_name(self, param):
		if isinstance(param, str_types):
			if param[:1] == "_":
				return param[1:]
			return param
		return param

	def __get_pam_scaled_name(self, param):
		param = self.__get_pam_name(param)
		if self.__default_scaled:
			return param
		return "_%s" % param

	def __get_pam_united_name(self, param):
		param = self.__get_pam_name(param)
		if not self.__default_scaled:
			return param
		return "_%s" % param

	def __get_pam_deps(self, param):
		try:
			return self.__cache_deps[param]
		except:
			if param not in self.__parameters:
				return []

			value = self.__parameters[param]
			if type(value) == types.FunctionType:
				self.__cache_deps[param] = list(map(self.__get_pam_name, self.__function_getargs(value)))
			else:
				self.__cache_deps[param] = []

			return self.__cache_deps[param]

	def __get_pam_sups(self, param):
		try:
			return self.__cache_sups[param]
		except:
			sups = []
			for param2 in self.__parameters:
				if param in self.__get_pam_deps(param2):
					sups.append(param2)

			if sups or param in self.__parameters:
				self.__cache_sups[param] = sups
			return sups

	############# PARAMETER RETRIEVAL ##########################################

	def __get(self, args, kwargs={}, default_scaled=None):
		'''
		Retrieve the parameters specified in args, with temporary values overriding
		defaults as in kwargs. Parameters are returned as Quantity's.
		'''
		self.__process_override(kwargs)

		arg_islist = type(args[0]) == list

		if len(args) == 1 and not arg_islist:
			result = self.__get_param(args[0], kwargs, default_scaled)
			if self.__parameters_bounds is not None:
				kwargs[args[0]] = result
				self.__forward_check_bounds(args, kwargs)
			return result

		if arg_islist:
			args = args[0]

		results = self.__get_params(args, kwargs, default_scaled)
		kwargs.update(results)
		if self.__parameters_bounds is not None:
			kwargs.update(results)
			self.__forward_check_bounds(args, kwargs)
		return results

	def __forward_check_bounds(self, args, kwargs):
		'''
		Check that bounds on parameters are not violated due to the changes
		specified in args and kwargs.
		'''
		checked = []
		for arg in args:
			if isinstance(arg, str_types):
				for pam in self.__get_pam_sups(arg):
					if pam in self.__parameters_bounds:
						keys = self.__get_pam_deps(pam)
						check = True
						for key in keys:
							if key not in kwargs and key not in self.__parameters:
								check = False
								break
						if check:
							self.__get_param(pam, kwargs)
						else:
							warnings.warn(errors.ParameterBoundsUncheckedWarning("Parameter '%s' might be outside bounds. Insufficient parameters passed to check." % pam))

	def __get_params(self, args, kwargs={}, default_scaled=None):
		rv = {}
		for arg in args:
			rv[self.__get_pam_name(arg)] = self.__get_param(arg, kwargs, default_scaled)
		return rv

	def __get_param(self, arg, kwargs={}, default_scaled=None):
		'''
		Returns the value of a param `arg` with its dependent variables overriden
		as in `kwargs`. If `arg` is instead a function, a string, or a Quantity, action is taken to
		evaluate it where possible.
		'''
		if arg == '_':
			raise ValueError()
		pam_name = self.__get_pam_name(arg)

		# If the parameter is actually a function or otherwise not directly in the dictionary of stored parameters
		if not isinstance(arg, str_types) or (pam_name not in kwargs and pam_name not in self.__parameters):
			return self.__eval(arg, kwargs, default_scaled)
		else:
			scaled = default_scaled if default_scaled is not None else self.__default_scaled
			if arg[:1] == "_":  # .startswith("_"):
				arg = arg[1:]
				scaled = not scaled

			# If the parameter is temporarily overridden, return the override value
			if arg in kwargs:
				return self.__get_quantity(kwargs[arg], param=arg, scaled=scaled)

			# If the parameter is a function, evaluate it with local parameter values (except where overridden in kwargs)
			elif type(self.__parameters[arg]) is types.FunctionType:
				return self.__get_quantity(self.__eval_function(arg, kwargs)[arg], param=arg, scaled=scaled)

			# Otherwise, return the value currently stored in the parameters
			else:
				if scaled:
					try:
						return self.__cache_scaled[arg]
					except:
						self.__cache_scaled[arg] = self.__get_quantity(self.__parameters[arg], param=arg, scaled=scaled)
						return self.__cache_scaled[arg]
				return self.__get_quantity(self.__parameters[arg], param=arg, scaled=scaled)

	def __process_override(self, kwargs, restrict=None):
		'''
		Process kwargs and make sure that if one of the provided overrides
		corresponds to an invertable function, that the affected variables are also included
		as overrides also. An warning is thrown if these variables are specified also
		and are inconsistent.
		'''

		if len(kwargs) == 0:
			return

		if restrict is None:
			restrict = list(kwargs.keys())

		if len(restrict) == 0:
			return

		def check_overspecified(param, kwargs, vals):
			return

		def pam_ordering(dependencies, pam_order=[]):
			'''
			This function returns parameter names in the following format:
			[ param_name, param_name, ...]
			Such that for any index, any parameters with greater index do
			not depend on parameters with index less than or equal to that index.
			'''
			new = set()
			for pam, deps in dependencies.items():
				if pam not in pam_order:
					for dep in deps:
						if dep not in pam_order and (dep not in dependencies or len(dependencies[dep].difference(set(pam_order)))):
							pam_order.append(dep)
							new.add(dep)
					if len(deps.difference(set(pam_order))) == 0:
						pam_order.append(pam)
						new.add(pam)

			if len(new) == 0 and len(set(dependencies.keys()).difference(set(pam_order))) != 0:
				raise ValueError("Function dependencies are circular.")
			elif len(new) == 0:
				return pam_order
			else:
				return pam_ordering(dependencies, pam_order)

		# Order overrides to avoid clash of functions
		dependencies = {}
		for pam in restrict:
			if pam[0] == "_":
				raise ValueError("Parameter type is autodetected when passed as a keyword argument. Do not use '_' to switch between scaled and unitted parameters.")
			val = kwargs[pam]
			if type(val) in str_types:
				val = self.__get_function(val)
				kwargs[pam] = val
			if type(val) is tuple and type(val[0]) is types.FunctionType:
				val = val[0]
			if type(val) is tuple and type(val[0]) in str_types:
				val = self.__get_function(val[0])
			if type(val) is types.FunctionType:
				pam = self.__get_pam_name(pam)
				deps = [self.__get_pam_name(dep) for dep in self.__function_getargs(val)]
				dependencies[pam] = set(deps)

		pam_order = pam_ordering(dependencies)

		# First evaluate functions to avoid errors later on
		for pam in pam_order:
			if pam in kwargs:
				val = kwargs[pam]
				new = kwargs.copy()
				del new[pam]
				kwargs[pam] = self.__get_param(val, new)

		# Now, ratify these changes through the parameter sets to ensure
		# that the effects of these overrides is properly implemented
		# inverting any methods with the provided overrides
		# and then recursing on any newly returned values.
		# If a method is not invertible, and it is request,
		# print a warning to this extent.
		new = {}
		for pam in restrict:
			if type(self.__parameters.get(pam)) is types.FunctionType:
				if pam in self.__get_pam_deps(pam):
					vals = self.__eval_function(pam, kwargs)
					for key in vals:
						if key in kwargs and self.__get_quantity(vals[key],scaled=True) != self.__get_quantity(kwargs[key],scaled=True) or key in new and self.__get_quantity(vals[key],scaled=True) != self.__get_quantity(new[key],scaled=True):
							raise errors.ParameterOverSpecifiedError("Parameter %s is overspecified, with contradictory values. (%s vs. %s)" % (key,vals[key],kwargs[key] if key in kwargs else new[key]) )
					new.update(vals)
				else:
					warnings.warn(errors.ParameterInconsistentWarning("Parameters are possibly inconsistent! The function representing '%s' was overridden because it was not invertable, and so the underlying variables (%s) have not been updated." % (pam, ','.join(self.__function_getargs(self.__parameters[pam])))))

		if len(new) != 0:
			kwargs.update(new)
			self.__process_override(kwargs, restrict=list(new.keys()))

	def __eval_function(self, param, kwargs={}):
		'''
		Returns a dictionary of parameter values. If the param variable itself is provided,
		then the function has its inverse operator evaluated. Functions must be of the form:
		def f(<pam>,<pam>,<param>=None)
		If <pam> is prefixed with a "_", then the scaled version of the parameter is sent through
		instead of the Quantity version.
		'''

		f = self.__parameters.get(param)
		deps = self.__get_pam_deps(param)
		deps_ = self.__function_getargs(f)

		# Check if we are inverting or evaluating (if param is in kwargs, we are inverting), and prepare.
		if param in kwargs:
			if deps[-1] != param:
				raise errors.ParameterNotInvertableError("Configuration requiring the inverting of a non-invertable map for %s." % param)
		else:
			if deps[-1] == param:
				deps = deps[:-1]
				deps_ = deps_[:-1]

		# Compute required arguments for functional argument
		params = self.__get_params(deps_, kwargs)
		args = [val for val in [params[self.__get_pam_name(x)] for x in deps_]]

		if param in kwargs: # Invert and return updated parameter values
			r = f(*args)
			if type(r) not in (list,tuple):
				r = (r,)

			inverse = {}

			for i, arg in enumerate(deps[:-1]): # Iterate through results except for final dep which is param or _param
				pam = self.__get_pam_name(arg)
				inverse[pam] = self.__get_quantity(r[i], param=pam)

			return inverse
		else: # Return value of function (from cache if possible)
			if param in self.__cache_funcs:
				cached = self.__cache_func_handler(param=param, params=args)
				if cached is not None:
					return {param: cached}
				else:
					value = f(*args)
					self.__cache_func_handler(param=param, value=value, params=args)
					return {param: value}
			else:
				return {param: f(*args)}


	def __cache_func_handler(self, param, value=None, params=None):
		'''
		Retrieve and set function cache.
		'''
		if param in self.__cache_funcs:
			if value is None:
				if self.__cache_funcs[param] is None:
					return None
				value, conditions = self.__cache_funcs[param]
				if conditions == params:
					return value
				else:
					return None
			else:
				self.__cache_funcs[param] = (value, params)

	def cache(self, **kwargs):
		'''
		cache(**kwargs)

		:param kwargs: Dictionary of boolean values
		:type kwargs: :class:`dict`

		A utility function to toggle caching of particular parameters. When
		cache is enabled, if a parameter function has been called before with the
		same parameter values, then it returns the old value. Note that only
		one set of parameters is remembered, and so this caching is designed
		only for situations where the parameter is not expected to change at
		every call.

		Example:

		>>> p.cache(x=True, y=False)

		This will enable caching for *x* and disable it for *y*.
		'''
		for kwarg, cache_on in kwargs.items():
			if kwarg in self.__cache_funcs and not cache_on:
				self.__cache_funcs.pop(kwarg)
			if kwarg not in self.__cache_funcs and cache_on:
				self.__cache_funcs[kwarg] = None

	def __get_quantity(self, value, param=None, unit=None, scaled=False):
		'''
		Return a Quantity or scaled float associated with the value provided
		and the dimensions of param.
		'''

		q = None

		t = type(value)

		if t is types.FunctionType:
			return value
		else:
			if scaled:

				# If tuple of (value,unit) is presented
				if t is tuple:
					if len(value) != 2:
						raise errors.QuantityCoercionError("Tuple specifications of quantities must be of form (<value>,<unit>). Was provided with %s ." % str(value))
					else:
						q = Quantity(value[0], value[1], dispenser=self.__units)
						q = q.value / self.__unit_scaling(q.units)

				elif isinstance(value, Quantity):
					q = value.value / self.__unit_scaling(value.units)

				else:  # if t in (float,complex,long,int,np.ndarray):
					q = value

			else:

				# If tuple of (value,unit) is presented
				if t is tuple:
					if len(value) != 2:
						raise errors.QuantityCoercionError("Tuple specifications of quantities must be of form (<value>,<unit>). Was provided with %s ." % str(value))
					else:
						q = Quantity(value[0], value[1], dispenser=self.__units)

				elif isinstance(value, Quantity):
					q = value

				else:  # if t in (float,complex,long,int,np.ndarray):
					if unit is None and param is None:
						unit = self.__get_unit('')
					elif unit is not None:
						unit = self.__get_unit(unit)
					else:
						unit = self.__get_unit(''  if self.__parameters_spec.get(param) is None else self.__parameters_spec.get(param))
					if isinstance(value, list):
						value = np.array(value)
					q = Quantity(value * self.__unit_scaling(unit), unit, dispenser=self.__units)

		if q is None:
			raise errors.QuantityValueError("Unknown value type '%s' with value: '%s'" % (t, value))

		if self.__parameters_bounds is not None and param is not None and param in self.__parameters_bounds:
			q = self.__check_bounds(self.__parameters_bounds[param], q)

		return q

	def __eval(self, arg, kwargs={}, default_scaled=None):

		if default_scaled is None:
			default_scaled = self.__default_scaled

		t = type(arg)

		if t == tuple:
			return self.__get_quantity((self.__eval(arg[0], kwargs), arg[1]), scaled=default_scaled)

		elif t == types.FunctionType:
			deps = self.__function_getargs(arg)
			params = self.__get_params(deps, kwargs)
			args = [val for val in [params[self.__get_pam_name(x)] for x in deps]]  # Done separately to avoid memory leak when cythoned.
			return arg(*args)

		elif isinstance(arg, Quantity):
			return self.__get_quantity(arg, scaled=default_scaled)

		elif isinstance(arg, str_types) or arg.__class__.__module__.startswith('sympy'):
			try:
				if isinstance(arg, str_types):
					# We have a string which cannot be a single parameter. Check to see if it is trying to be.
					arg = sympy.S(arg, sympy.abc._clash)
					fs = list(arg.free_symbols)
					if len(fs) == 1 and str(arg) == str(fs[0]):
						raise errors.ParameterInvalidError("There is no parameter, and no interpretation, of '%s' which is recognised by Parameters." % arg)
				return self.__eval(self.optimise(arg), kwargs=kwargs, default_scaled=default_scaled)
			except errors.ParameterInvalidError as e:
				raise e
			except Exception as e:
				raise errors.SymbolicEvaluationError("Error evaluating symbolic statement '%s'. The message from SymPy was: `%s`." % (arg, e))

		elif isinstance(arg, (complex, int, float, long)):
			return arg

		raise errors.ParameterInvalidError("There is no parameter, and no interpretation, of '%s' which is recognised by Parameters." % arg)

	################ SET PARAMETERS ############################################

	def __is_valid_param(self, param, allow_leading_underscore=True):
		return re.match("^[%sA-Za-z][_a-zA-Z0-9]*$" % ('_' if allow_leading_underscore else ''), param)

	def __check_valid_params(self, params, allow_leading_underscore=True):
		bad = []
		for param in params:
			if not self.__is_valid_param(param, allow_leading_underscore=allow_leading_underscore):
				bad.append(param)
		if len(bad) > 0:
			raise errors.ParameterInvalidError("Attempt to set invalid parameters: %s . Parameters must be valid python identifiers matching ^[%sA-Za-z][_a-zA-Z0-9]*$." % (','.join(bad), '_' if allow_leading_underscore else ''))

	def __set(self, kwargs):

		self.__cache_deps = {}
		self.__cache_sups = {}

		self.__check_valid_params(kwargs, allow_leading_underscore=False)

		for param, val in kwargs.items():
			if param in self.__cache_funcs:
				self.__cache_funcs[param] = None
			if param in self.__cache_scaled:  # Clear cache if present.
				del self.__cache_scaled[param]
			if isinstance(val, (types.FunctionType,) + str_types):
				self.__parameters[param] = self.__check_function(param, self.__get_function(val))
				self.__spec({param: self.__get_unit('')})
			elif isinstance(val, (list, tuple)) and isinstance(val[0], (types.FunctionType,) + str_types):
				self.__parameters[param] = self.__check_function(param, self.__get_function(val[0]))
				self.__spec({param: self.__get_unit(val[1])})
			else:
				self.__parameters[param] = self.__get_quantity(val, param=param)
				if isinstance(self.__parameters[param], Quantity):
					self.__spec({param: self.__parameters[param].units})
			if param in dir(type(self)):
				warnings.warn(errors.ParameterNameWarning("Parameter '%s' will not be accessible using the attribute notation `p.%s`, as it conflicts with a method name of Parameters." % (param, param)))

	def __update(self, kwargs):

		self.__check_valid_params(kwargs)

		self.__process_override(kwargs)

		for param, value in kwargs.items():
			if param not in self.__parameters or not (isinstance(self.__parameters.get(param), types.FunctionType) and param in self.__get_pam_deps(param)):
				self.__set({param: kwargs[param]})

	def __and__(self, other):
		if not isinstance(other, dict):
			raise errors.ParametersException("The binary and operator is used to set the unit specification for parameters; and requires a dictionary of units.")
		for param, units in other.items():
			if isinstance(self.__parameters.get(param), Quantity):
				self.__parameters[self.__get_pam_name(param)] = self.__get_param(self.__get_pam_united_name(param))(units)
		self.__spec(other)

	def __spec(self, kwargs):
		''' Set units for parameters. '''
		for arg in kwargs:
			self.__parameters_spec[arg] = self.__get_unit(kwargs[arg])
			if self.__parameters.get(arg) is not None:
				self.__parameters[arg].units = self.__parameters_spec[arg]

	def __remove(self, param):
		if param in self.__parameters:
			del self.__parameters[param]
		if param in self.__parameters_spec:
			del self.__parameters_spec[param]

	def forget(self, *params):
		'''
		forget(*params)

		:param params: List of parameter names to forget.
		:type params: :class:`tuple` of :class:`str`

		:returns: A reference to the parent :class:`Parameters` instance.

		This is the way to remove parameters that are stored inside a :class:`Parameters`
		instance.

		Example:

		>>> p.forget('x','y','z')

		This removes parameters *x*, *y* and *z* from the parameter list.
		'''
		for param in params:
			self.__remove(param)
		return self

	def __sympy_to_function(self, expr):
		try:
			expr = sympy.S(expr, locals=sympy.abc._clash)
			syms = list(expr.free_symbols)
			f = sympy.utilities.lambdify(syms, expr, dummify=False, modules=['numpy','mpmath','math','sympy'])
			return f
		except Exception, e:
			print(e)
			raise errors.SymbolicEvaluationError('String \'%s\' is not a valid symbolic expression.' % (expr))

	def __get_function(self, expr):
		if isinstance(expr, types.FunctionType):
			return expr
		return self.__sympy_to_function(expr)

	def __check_function(self, param, f, forbidden=None):

		_param = '_' + param

		inspection = inspect.getargspec(f)
		if inspection.varargs is not None or inspection.keywords is not None:
			raise ValueError("Cannot add parameter function that uses varargs or keyword arguments for '%s'." % param)
		if (param in inspection.args and inspection.args.index(param) != len(inspection.args) - 1) or (_param in inspection.args and inspection.args.index(_param) != len(inspection.args) - 1):
			raise ValueError("Self-reference for inversion must be the last parameter provided in args for '%s'." % param)
		if (param in inspection.args or _param in inspection.args)and inspection.defaults != (None,):
			raise ValueError("Cannot add parameter function that does not set a default value of None for self-referential parameter in definition for '%s'." % param)
		if param not in inspection.args and _param not in inspection.args and inspection.defaults != None:
			raise ValueError("Cannot add parameter function that provides default values for parameters in '%s'." % param)

		args = list(self.__function_getargs(f))

		if param in args:
			args.remove(param)

		if forbidden is None:
			forbidden = []
		else:
			for arg in forbidden:
				if arg in args:
					raise errors.ParameterRecursionError("Adding function would result in recursion with function '%s'" % arg)
		forbidden.append(param)

		for arg in args:
			if isinstance(self.__parameters.get(arg, None), types.FunctionType):
				self.__check_function(arg, self.__parameters.get(arg), forbidden=forbidden[:])

		return f

	def __function_getargs(self, f):  # faster than inspect.getargspec(f).args
		return f.__code__.co_varnames[:f.__code__.co_argcount]

	def __basis_scale(self, unit):
		unit = self.__get_unit(unit)
		scaling = Quantity(1, None, dispenser=self.__units)

		for dim, power in unit.dimensions.items():
			scaling *= self.__scalings.get(dim, Quantity(1, self.__units.basis()[dim], dispenser=self.__units)) ** power

		return scaling

	def __unit_scaling(self, unit):
		'''
		Returns the float that corresponds to the relative scaling of the
		provided unit compared to the intrinsic scaling basis of the parameters.

		dimensionless value = quantity / unit_scaling = quantity * unit_scale / basis_scale
		'''

		if unit in self.__scaling_cache:
			return self.__scaling_cache[unit]

		scale = self.__basis_scale(unit)
		scaling = scale.value * scale.units.scale(unit)

		self.__scaling_cache[unit] = scaling
		return scaling

	################ EXPOSE PARAMETERS #########################################
	def __call__(self, *args, **kwargs):
		if args:
			return self.__get(args, kwargs)

		self.__update(kwargs)
		return self

	def _(self, *args, **kwargs):
		if args:
			return self.__get(args, kwargs, not self.__default_scaled)

		self.__update(kwargs)
		return self

	def __getattr__(self, name):
		if name[:2] == "__" or name[:11] == "_Parameters":
			raise AttributeError()
		return self.__get_param(name)

	def __setattr__(self, attr, value):
		if attr.startswith('__') or attr.startswith('_Parameters'):
			return super(Parameters, self).__setattr__(attr, value)
		return self.__set({attr: value})

	def __lshift__(self, other):

		if not isinstance(other, dict):
			raise errors.ParametersException("The left shift operator sets parameter values without interpretation; such as functions. It accepts a dictionary of parameter values.")

		self.__set(other)
		return self

	def __dir__(self):
		res = dir(type(self)) + list(self.__dict__.keys())
		res.extend(self.__parameters.keys())
		return res

	################# Other Magic ##############################################
	def __repr__(self):
		return "< Parameters with %d definitions >" % len(self.__parameters)

	def __len__(self):
		return len(self.__parameters)

	def __iter__(self):
		params = sorted(self.__parameters.keys())
		for param in params:
			yield param

	def __getitem__(self, key):
		if type(key) == int:
			return sorted(self.__parameters.keys())[key]
		return self.__get(key)

	def __setitem__(self, key, value):
		self.__update({key: value})

	################# Show parameters ########################################

	def __table(self, table):

		def text_len(text):
			if '\033[' in text:
				return len(text) - 11
			return len(text)

		def column_width(i, text):
			if '\033[' in text:
				return col_width[i] + 11
			return col_width[i]

		col_width = [max(text_len(x) for x in col) for col in zip(*table)]
		output = []
		for line in table:
			output.append("| " + " | ".join("{:^{}}".format(x, column_width(i, x)) for i, x in enumerate(line)) + " |")

		return '\n'.join(output)

	def show(self):
		'''
		show(self)

		This method simply prints a formatted table of parameters stored in the
		:class:`Parameters` instance. It shows the parameter and its dependencies,
		its united value, and its non-dimensionalised value.

		To use it, simply run:

		>>> p.show()
		'''
		if len(self.__parameters) == 0:
			return 'No parameters have been specified.'

		parameters = [[colour_text('Parameter', 'WHITE', True), colour_text('Value', 'WHITE', True), colour_text('Scaled', 'WHITE', True)]]
		for param in sorted(self.__parameters.keys()):

			if self.__default_scaled:
				key_scaled, key = param, '_%s' % param
			else:
				key_scaled, key = '_%s' % param, param

			if isinstance(self.__parameters[param], types.FunctionType):
				v = 'Unknown'
				vs = 'Unknown'
				try:
					v = str(self.__get_param(key))
					vs = str(self.__get_param(key_scaled))
				except:
					pass
				parameters.append([
					'%s(%s)' % (param, ','.join(self.__get_pam_deps(param))),
					v,
					vs])

			else:
				parameters.append([param, str(self.__get_param(key)), str(self.__get_param(key_scaled))])

		for param in sorted(self.__parameters_spec.keys()):
			if param not in self.__parameters:
				parameters.append([colour_text(param, 'CYAN'), colour_text("- %s" % self.__parameters_spec[param], 'CYAN'), colour_text("-", 'CYAN')])

		print(self.__table(parameters))

	################## PARAMETER BOUNDS ####################################

	def bounds(self, *params, **bounds):
		'''
		bounds(*params, **bounds)

		:param params: Sequence of parameter names for which to query the bounds.
		:type param: tuple
		:param bounds: Dictionary specifying parameter bounds to be applied.
		:type bounds: dict

		:returns: Dictionary of :class:`Bound` objects, or, if only one parameter is specified, then just one :class:`Bound` object. If no bounds exist for a parameter, :python:`None` is returned.

		This method allows you to simply query and set bounds on parameters. For
		more advanced bounds setting, you should use :func:`set_bounds`. For each
		parameter, valid bounds are specified by a two-tuple, or a list of two-tuples.
		If one of the extremum values is None, it is set to -infinity or +infinity, depending upon
		whether it is the upper or lower bound.

		For example:

		>>> p.bounds(x=(0,100)) # Bounds x between 0 and 100 inclusive.

		>>> p.bounds(y=[(0,50),(150,200)]) # Bounds y between 0-50 or between 150-200.

		>>> p.bounds('x') # Returns the Bounds object associated with x

		>>> p.bounds('x',x=(0,None)) # Sets the bounds on x to be [0,+inf], and then returns the :class:`Bounds` object associated with x.

		Note that multiple parameters can be queried and set at the same time.

		.. warning:: Using parameter bounds greatly increases the amount of computation done during parameter retrieval. It is recommended that you do not use parameter bounds in contexts which require minimal runtime, such as numerical integration.
		'''
		self.set_bounds(bounds)

		if len(params) > 0:
			use_dict = type(params[0]) == list
			if use_dict:
				params = params[0]

			bounds = {}
			for param in params:
				bounds[param] = self.__parameters_bounds[param].bounds if param in self.__parameters_bounds else None
			if not use_dict and len(bounds) == 1:
				return bounds[list(bounds.keys())[0]]
			return bounds

	def set_bounds(self, bounds_dict, error=True, clip=False, inclusive=True):
		'''
		set_bounds(bounds_dict, error=True, clip=False, inclusive=True)

		:param bounds_dict: A dictionary with parameters as keys with a valid bounding specification (described below).
		:type bounds_dict: dict
		:param error: :python:`True` if a parameter found to be outside specified bounds throw an error; or if :python:`clip` is :python:`True`, whether a warning should be generated. :python:`False` otherwise.
		:type error: bool
		:param clip: :python:`True` if the relevant parameter should be clipped to the nearest bound edge (assumes inclusive is True). :python:`False` otherwise.
		:type clip: bool
		:param inclusive: :python:`True` if the upper and lower bounds should be included in range. :python:`False` otherwise.
		:type inclusive: bool

		This method provides you with greater control than the shorthand methods
		described in the class documentation. As a reminder the shorthand methods
		looked like:

		>>> p[’x’] = (0,100)

		If one of the extremum values is None, it is set to -infinity or +infinity, depending upon
		whether it is the upper or lower bound. If a disjointed bound is necessary, you can use:

		>>> p[’x’] = [ (None,10), (15,None) ]

		This method is much more flexible; for example:

		>>> p.set_bounds( {'x':(0,100)}, error=True, clip=True, inclusive=True )

		Will warp 'x' to the closer of 0 or 100 if outside of the range [0,100],
		reporting a warning in the process.

		.. warning:: Using parameter bounds greatly increases the amount of computation done during parameter retrieval. It is recommended that you do not use parameter bounds in contexts which require minimal runtime, such as numerical integration.
		'''
		if not isinstance(bounds_dict, dict):
			raise ValueError("Bounds must be specified as a dictionary. Provided with: '%s'." % (bounds_dict))
		for key, bounds in bounds_dict.items():
			if not isinstance(bounds, list):
				bounds = [bounds]
			bounds_new = []
			for bound in bounds:
				if not isinstance(bound, tuple) or len(bound) != 2:
					raise ValueError("Bounds must be of type 2-tuple. Received '%s'." % (bound))

				lower = bound[0]
				if lower is None:
					lower = (-np.inf, self.units(key))
				lower = self.__get_quantity(lower, param=key)

				upper = bound[1]
				if upper is None:
					upper = (np.inf, self.units(key))
				upper = self.__get_quantity(upper, param=key)
				bounds_new.append((lower, upper))

			if self.__parameters_bounds is None:
				self.__parameters_bounds = {}
			self.__parameters_bounds[key] = Bounds(key, self.units(key), bounds_new, error=error, clip=clip, inclusive=inclusive)

	def __check_bounds(self, bounds, value):
		if isinstance(value, Quantity):
			scaled = False
			value_comp = value
		else:
			scaled = True
			value_comp = Quantity(value, self.__parameters_spec.get(bounds.param, None), dispenser=self.__units)

		if bounds.inclusive:
			for bound in bounds.bounds:
				if bound[0] <= value_comp and bound[1] >= value_comp:
					return value
		else:
			for bound in bounds.bounds:
				if bound[0] < value_comp and bound[1] > value_comp:
					return value

		if bounds.clip:
			if bounds.error:
				warnings.warn(errors.ParameterOutsideBoundsWarning("Value %s for '%s' outside of bounds %s. Clipping to nearest allowable value." % (value, bounds.param, bounds.bounds)))
			blist = []
			for bound in bounds.bounds:
				blist.extend(bound)
			dlist = map(lambda x: abs(x - value_comp), blist)
			v = np.array(blist)[np.where(dlist == np.min(dlist))]
			if scaled:
				return v
			else:
				return v.value / self.__unit_scaling(v.unit)
		elif bounds.error:
			raise errors.ParameterOutsideBoundsError("Value %s for '%s' outside of bounds %s" % (value, bounds.param, bounds.bounds))

		warnings.warn(errors.ParameterOutsideBoundsWarning("Value %s for '%s' outside of bounds %s. Using value anyway." % (value, bounds.param, bounds.bounds)))
		return value

	################## RANGE UTILITY #######################################

	def range(self, *args, **ranges):
		'''
		range(*args, **ranges)

		:param args: A sequence of parameters (or parameter expressions).
		:type args: tuple
		:param ranges: A dictionary of overrides and range specifications.
		:type ranges: dict

		:returns: A sequence of parameter values if there is a single parameter requested and it is not enclosed in a list, and dictionary of values otherwise.

		This method provides a solution to the common problem of iterating over
		parameter ranges, or investigating how one parameter changes as a function
		of another. It has a similar syntax for parameter extraction by calling
		the :class:`Parameter` instance, and indeed provides a superset of the
		functionality. It is kept separate for peformance considerations.

		One can think about the syntax of this method as overriding the parameter
		value with a sequence of values, rather than a specific value. For example:

		>>> p << {'y':lamdba x:x**2}
		>>> p.range( 'y', x = [0,1,2,3] )
		[0,1.,4.,9.]

		In this simple example, we see that we can iterate over a provided array of
		values. Arrays may be input as lists or numpy ndarrays; and returned arrays are typically numpy arrays.

		The values for parameter overrides can also be provided in a more abstract notation; such that the
		range will be generated when the function is called. Parameters accepts ranges in the following
		forms:

			- (*<start>*, *<stop>*, *<count>*) : which will generate a linear array from <start> to <stop> with <count> values.
			- (*<start>*, *<stop>*, *<count>*, ..., *<sampler>*) : which is as above, but where the <sampler> is expected to generate the array. <sampler> can be a string (either ‘linear’,’log’,’invlog’ for linear, logarithmic, or inverse logarithmic distributions respectively); or a function which takes arguments <start>, <stop>, <count> and any other arguments from "...". Note that when you specify your own function, the <start>, <stop> and <count> variables need not be interpreted as their name suggests.

		Example:

		>>> p.range( 'y', x = (0,10,2) )
		[0.,100.]

		It is also possible to determine multiple parameters at once.

		>>> p.range( ‘x’, ‘y’, x=(0,10,2) )
		{'x':[0.,10.], 'y':[0.,100.]}

		If multiple overrides are provided, they must either be constant or have the same length.

		>>> p.range('x', x=(0,10,2), z=1 ) # This is OKAY

		>>> p.range('x', x=(0,10,2), z=[3,4]) # This is also OKAY

		>>> p.range( 'x', x=(0,10,2), z=[1,2,3] ) # This is NOT okay.
		'''

		if len(args) == 0:
			raise ValueError('Please specify output variables from ranges.')

		values = None
		static = {}
		lists = {}

		# Separate out static coefficients
		for param, pam_range in ranges.items():
			if not isinstance(pam_range, (list, np.ndarray, tuple)) or type(pam_range) is tuple and len(pam_range) < 3:  # Using negated check to maximise allowed input types.
				static[param] = pam_range

		# Generate sequences in the context of static coefficients
		# Note: It is not necessary to worry about clashes at this
		#       stage. They will be detected in the self.__get() method.
		count = None
		for param, pam_range in ranges.items():
			pam_range = self.__range_interpret(param, pam_range, params=static)
			if isinstance(pam_range, (list, np.ndarray)):
				lists[param] = pam_range
				count = len(pam_range) if count is None else count
				if count != len(pam_range):
					raise ValueError("Not all parameters have the same range")

		if count is None:
			return self.__get(args, ranges)

		if type(args[0]) == list:
			pargs = args[0]
		else:
			pargs = args

		for i in range(count):
			d = {}
			d.update(static)
			for key in lists:
				d[key] = lists[key][i]

			argvs = self.__get(args, d)

			if type(argvs) == dict:
				if values is None:
					values = {}
				for arg in argvs:
					if arg not in values:
						values[arg] = []
					values[arg].append(argvs[arg])
			else:
				if values is None:
					values = []
				values.append(argvs)

		return values

	def __range_sampler(self, sampler):
		if isinstance(sampler, str_types):
			if sampler == 'linear':
				return np.linspace
			elif sampler == 'log':
				def logspace(start, end, count):
					logged = np.logspace(1, 10, count)
					return (logged - logged[0]) * (end - start) / logged[-1] + start
				return logspace
			elif sampler == 'invlog':
				def logspace(start, end, count):
					logged = np.logspace(1, 10, count)
					return (logged[::-1] - logged[0]) * (end - start) / logged[-1] + start
				return logspace
			else:
				raise ValueError("Unknown sampler: %s" % sampler)
		elif type(sampler) == types.FunctionType:
			return sampler
		else:
			raise ValueError("Unknown type for sampler: %s" % type(sampler))

	def __range_interpret(self, param, pam_range, params=None):
		if isinstance(pam_range, tuple) and len(pam_range) >= 3:

			if len(pam_range) >= 4:  # Then assume format (*args, sampler), with sampler(*args) being the final result.
				args = list(pam_range[:-1])
				sampler = pam_range[-1]
			elif len(pam_range) == 3:  # Then assume format (start, end, count)
				args = list(pam_range)
				sampler = 'linear'
			else:
				raise ValueError("Unknown range specification format: %s." % pam_range)

			sampler = self.__range_sampler(sampler)

			for i, arg in enumerate(args):
				if isinstance(arg, (tuple, Quantity) + str_types):
					pars = {param: arg}
					if type(params) is dict:
						pars.update(params)
					args[i] = self.__get([self.__get_pam_scaled_name(param)], pars)

			# Note: param keyword cannot appear in params without keyword repetition in self.range.
			return sampler(*args)
		return pam_range

	################## Function iteration ##################################
	def ranges_iterator(self, ranges, params={}, masks=None, function=None, param_args=(), function_args=(), function_kwargs={}, nprocs=None, distributed=False, ranges_eval=None, progress=True):
		'''
		ranges_iterator(ranges, params={}, masks=None, function=None, param_args=(), function_args=(), function_kwargs={}, nprocs=None, distributed=False, ranges_eval=None, progress=True)

		This method is shorthand for:

		>>> RangesIterator(parameters=self, ranges=ranges, params=params, masks=masks, function=function, \
						function_kwargs=function_kwargs, nprocs=nprocs, ranges_eval=ranges_eval, progress=progress)

		The :class:`RangesIterator` object allows you to iterate over nested parameter
		ranges, which is useful when you want to sweep out a multidimensional parameter
		space; for example, when you want to make a 2D plot of some function. In
		some sense, this is a generalisation of the Python :func:`map` function.

		For more information, please refer to the :class:`RangesIterator` documentation.
		'''
		return RangesIterator(parameters=self, ranges=ranges, params=params, masks=masks, function=function, function_args=function_args, \
						function_kwargs=function_kwargs, nprocs=nprocs, distributed=distributed, ranges_eval=ranges_eval, progress=progress)

	################## CONVERT UTILITY #####################################
	def asvalue(self, **kwargs):
		'''
		asvalue(**kwargs)

		:param kwargs: A dictionary of parameter values
		:returns: Number (normally float, but could be complex, etc)

		A utility function to return what the united value of a parameter would
		be if it were overridden with *kwargs*. It is the logical partner
		of *asscaled(\*\*kwargs)*. For example:

		>>> p.asvalue(x=(1, 'nm'))
		1

		If no units are passed, then inout value is assumed to be non-dimensional,
		and units are lifted from the underlying parameter. For example:

		>>> p.x = (1,'ms')
		>>> p.asvalue(x=1)
		1000

		If multiple parameter values are specified, a dictionary of values is
		returned:
		>>> p.asvalue(x=1, y=2)
		{'x': 1000, 'y': 2}
		'''
		d = {}
		for param, value in kwargs.items():
			d[param] = self.convert(value, output=self.units(param), value=True)
		if len(d) == 1:
			return list(d.values())[0]
		return d

	def asscaled(self, **kwargs):
		'''
		asscaled(**kwargs)

		:param kwargs: A dictionary of parameter values
		:returns: Number (normally float, but could be complex, etc)

		A utility function to return what the scaled value of a parameter would
		be if it were overridden with *kwargs*. It is the logical partner
		of *asvalue(\*\*kwargs)*. For example:

		>>> p.asscaled(x=(1, 'm'))
		1

		Note that this is equivalent to :python:`p.convert((1, 'm'))`.
		If multiple parameter values are specified, a dictionary of values is
		returned:

		>>> p.asscaled(x=1, y=2)
		{'x': 1, 'y': 2}
		'''
		d = {}
		for param, value in kwargs.items():
			d[param] = self.convert(value)
		if len(d) == 1:
			return list(d.values())[0]
		return d

	def units(self, *params):
		'''
		units(*params)

		:param params: A sequence of parameters for which to extract the default units.
		:param type: tuple of str

		:returns: An dictionary of :class:`Units`, or, if only one param has been
		requested and not wrapped in a list, a single :class:`Units` object.

		This method returns the default units associated with a particular parameter.
		For example:

		>>> p << {'x': (1,'ms'), 'y': (1,'m')}
		>>> p.units('x')
		ms
		>>> p.units(['x'])
		{'x': ms}
		>>> p.units('x','y')
		{'x': ms, 'y': m}
		'''
		use_dict = False
		if len(params) == 1 and type(params[0]) == list:
			params = params[0]
			use_dict = True

		units = {}
		for param in params:
			units[param] = self.__parameters_spec.get(param, None)

		if len(params)==1 and not use_dict:
			return units[list(units.keys())[0]]
		return units

	def convert(self, quantity, input=None, output=None, value=True):
		'''
		convert(self,quantity,input=None, ouput=None, value=True)

		:param quantity: The quantity to be converted.
		:type quantity: :class:`Quantity`, `Quantity` tuple representation or any pythonic numeric type (including numpy arrays).
		:param input: The units of the inputed quantity (ignored if input type is :class:`Quantity`).
		:type input: :class:`None`, :class:`str`, or :class:`Units`
		:param output: The units to convert toward.
		:type output: :class:`None`, :class:`str`, or :class:`Units`
		:param value: Whether the function should return only the value (rather than the full :class:`Quantity` object).
		:type value: :class:`bool`

		:returns: Pythonic number if :python:`value` is :python:`True`, and :class:`Quantity` otherwise.
		'''

		if type(quantity) == list:
			return map(lambda q: self.convert(q, input, output, value), quantity)

		if type(quantity) == tuple and len(quantity) == 2:
			input = str(quantity[1])
			quantity = quantity[0]
		elif isinstance(quantity, Quantity):
			input = str(quantity.units)
			quantity = quantity.value

		if input is not None and output is not None:
			quantity /= self.__units(output).scale(input)

		elif input is not None:
			quantity /= self.__unit_scaling(self.__units(input))

		elif output is not None:
			quantity *= self.__unit_scaling(self.__units(output))

		if value:
			return quantity
		return Quantity(quantity, output, dispenser=self.__units)

	def optimise(self, param, *wrt, **params):
		'''
		optimise(param, *wrt, **params)

		This method returns either a function or string depending on whether
		the input :class:`param` consisted of more than a single parameter. For
		symbolic expressions, this can greatly speed up parameter retrieval. A
		similar mechanism is used internally to make parameter lookups fast. Additionally,
		it can pre-evaluate parameters that do not depend upon a parameter listed
		in `wrt`, subject to the parameter overrides of `params`. If `wrt` is not provided,
		only the functionalisation of the string representation of an expression is performed.

		:param param: Any parameter specification that is accepted by parameter retrieval.
		:param type: object
		:param wrt: Parameters for which all dependees should be preserved as variables.
		:type wrt: tuple
		:param params: Parameter overrides to use for the check as to whether a parameter
			is dependent on one of the parameters in `wrt`.
		:type params: dict

		:returns: A python function that can be passed to a Parameters instance for Parameters retrieval, or a string if the :class:`param` consisted of a single parameter name.
		:raises: ExpressionOptimisationError

		For example:

		>>> p.optimise('sin(x)*exp(-t)')
		< function with arguments x and t >

		>>> p.optimise('sin(x)*exp(-t)','t',x=1)
		<function with argument t, with x evaluated to 1>
		'''

		if param is None or isinstance(param, types.FunctionType) or isinstance(param, str_types) and self.__is_valid_param(param):
			return param

		elif isinstance(param, str_types) or type(param).__module__.startswith('sympy'):
			if len(wrt) > 0:
				subs = {}
				expr = sympy.S(param, locals=sympy.abc._clash)
				for symbol in expr.free_symbols:
					symbol = str(symbol)
					if symbol in self and self.is_constant(symbol, *wrt, **params):
						subs[symbol] = self.__get(str(symbol), params)
				expr = expr.subs(subs)
			else:
				expr = param
			return self.__sympy_to_function(expr)

		raise errors.ExpressionOptimisationError("No way to optimise parameter expression: %s ." % param)

	def is_resolvable(self, *args, **params):
		'''
		is_resolvable(*args, **params)

		:param args: Sequence of parameter names.
		:type args: tuple
		:param params: Dictionary of parameter value overrides.
		:type params: dict

		:returns: :python:`True` if each of the requested parameters can be
		successfully evaluated. :python:`False` otherwise.

		This method actually goes through
		the process of evaluating the parameter, so if you need its value, it is probably better to use a
		try-except block in your code around the usual parameter extraction code.

		Example:

		>>> p.x = 1
		>>> p.y = 'x*z'
		>>> p.is_resolvable('x')
		True
		>>> p.is_resolvable('x','y')
		False
		>>> p.is_resolvable('x','y',z=1)
		True
		'''
		try:
			self(*args, **params)
			return True
		except:
			return False

	def is_function(self, param, **params):
		'''
		is_function(param, **params)

		:param param: Name of parameter
		:type param: str
		:param params: Dictionary of parameter value overrides.
		:type params: dict

		:returns: :python:`True` if specified parameter is a function. :python:`False` otherwise.

		This method checks if a given parameter is a function, which usually
		implies it is dependent on other parameters. For example:

		>>> p.x = 3
		>>> p.is_function('x')
		False
		>>> p.is_function('x',x='3*t')
		True
		'''
		param_name = self.__get_pam_name(param)
		param_val = None

		if param_name in params:
			param_val = params[param]
		elif param_name in self.__parameters:
			param_val = self.__parameters[param]
		else:
			try:
				symbols = sympy.S(param, sympy.abc._clash).free_symbols
				for symbol in symbols:
					if str(symbol) != str(param):
						return True
			except:
				raise errors.ParameterInvalidError("This parameters instance has no parameter named '%s', and none was provided. Parameter may or may not be constant." % param)

		if isinstance(param_val, types.FunctionType) or isinstance(param_val, str_types) and isinstance(self.optimise(param_val), (types.FunctionType,) + str_types):
			return True
		return False

	def is_constant(self, param, *wrt, **params):
		'''
		is_constant(*args, **params)

		:param param: The param for which to test constancy.
		:type param: str
		:param wrt: A sequence of parameter names, with respect to which `param` should be constant.
		:type wrt: tuple
		:param params: Dictionary of parameter value overrides.
		:type params: dict

		:returns: :python:`True` if first listed parameter is independent of all subsequent parameters. :python:`False` otherwise.

		This method is useful for simplifying parameters in contexts where parameters
		are going to be often polled, by checking to see if they can be statically
		cached in your application. For example:

		>>> p.x = "3*t"
		>>> p.is_constant('x','t')
		False
		>>> p.is_constant('x','t',x=1)
		True
		'''
		if len(wrt) == 0:
			return True
		if param in wrt:
			return False

		param_name = self.__get_pam_name(param)
		param_val = None

		if param_name in params:
			param_val = params[param]
		elif param_name in self.__parameters:
			param_val = self.__parameters[param]
		else:
			try:
				symbols = sympy.S(param, sympy.abc._clash).free_symbols
				for symbol in symbols:
					if str(symbol) != str(param_val):
						if not self.is_constant(str(symbol), *wrt, **params):
							return False
				return True
			except errors.ParameterInvalidError as e:
				raise e
			except:
				raise errors.ParameterInvalidError("This parameters instance has no parameter named '%s', and none was provided. Parameter may or may not be constant." % param)

		if isinstance(param_val, str_types):
			param_val = self.optimise(param_val)
		else:
			param_val = self.__get_quantity(param_val)

		if isinstance(param_val, Quantity):
			return True
		elif isinstance(param_val, types.FunctionType):
			deps = self.__function_getargs(param_val)
			for dep in deps:
				dep = self.__get_pam_name(dep)
				if dep == param:
					continue
				if dep in wrt or not self.is_constant(dep, *wrt, **params):
					return False
			return True
		else:
			raise ValueError("Unable to check whether parameter '%s' of type '%s' is constant." % (param_val, type(param_val)))

	################## PLOTTING INTROSPECTION ##############################

	def plot(self, *params, **ranges):
		'''
		plot(*params, **ranges)

		:param params: A sequence of parameter expressions to plot.
		:type params: tuple
		:param ranges: A dictionary of parameter ranges and overrides.
		:type ranges: dict

		If Matplotlib is installed, this method provides a simple way to debug
		whether your parameter values are working as expected. The format of the
		ranges can be anything accepted by :func:`range`, but must only have one
		varying independent parameter.

		For example:

		>>> p.plot('x', x="sin(w*t)", w=2, t=(0,'2*3.142/w',50))

		The above example will generate a sinusoidal curve *x* vs. *t* plot
		at 50 equidistant intervals between 0 and the end of the first period.
		'''
		try:
			import matplotlib.pyplot as plt
		except:
			print(colour_text("Matplotlib is required for plotting features.", "RED", True))
			return

		if len(params) == 0:
			raise ValueError("You must specify a parameter to plot. i.e. plot('x',t=(0,10,10))")

		indep_count = 0
		for param, range in ranges.items():
				if type(range) == list or type(range) == tuple and len(range) > 2:
					indep_count += 1
					if indep_count > 1:
						raise ValueError("Plotting currently only supports one independent parameter.")
					indep = self.__get_pam_scaled_name(param)
					indep_units = self.units(indep)
		if indep_count == 0:
			raise ValueError("You must provide at least one range to act as independent parameter.")

		r = self.range(indep, *map(self.__get_pam_scaled_name, params), **ranges)

		plt.figure()
		for param in params:
			param_units = self.units(param)
			plt.plot(
					self.asvalue(**{indep: r[indep]}),
					self.asvalue(**{param: r[param]}),
					label=u"$%s\,(%s)$" % (self.__plot_subscripts(param), self.__plot_subscripts(unicode(param_units)) if param_units is not None else "units")
					)

		plt.xlabel(u"$%s\,(%s)$" % (self.__plot_subscripts(indep), self.__plot_subscripts(unicode(indep_units)) if indep_units is not None else 'units'))
		plt.legend(loc=0)

		plt.show()

	def __plot_subscripts(self, text):
		s = text.split('_')
		return '_{'.join(s) + '}' * (len(s) - 1)

	################## LOAD / SAVE PROFILES ################################

	@classmethod
	def load(cls, filename, **kwargs):
		'''
		load(cls, filename, **kwargs)

		:param filename: Filename from which to load parameters.
		:type filename: str
		:param kwargs: Dictionary of arguments to pass to :class:`Parameters` constructor.
		:type kwargs: dict

		:returns: A new :class:`Parameters` instance preloaded with the configuration contained in :python:`filename`.

		This is the method you should use to load a saved :class:`Parameters`
		configuration. For example:

		>>> p.load('params.py')

		The file being loaded should be a valid Python file, with one or more of
		the following variables available in the global namespace:
			- :python:`parameters` : a dictionary of parameter values with names as keys.
			- :python:`parameters_cache` : a dictionary of boolean values with names as keys (and where True indicates that the parameter should be cached, see :func:`cache`).
			- :python:`parameters_units` : a dictionary of parameter units with names as keys (only necessary to specify units for parameters which do not have a value attached to them, but for which it is useful to have default units)
			- :python:`dimension_scalings` : a dictionary of scalings with dimensions as keys (for valid scalings, see :func:`scaling`).
			- :python:`units_custom` : a list of dictionaries which contain the kwargs necessary to construct the custom unit (seel :func:`add_unit`).
		'''
		profile = imp.load_source('profile', filename)

		p = cls(**kwargs)

		p.scaling(**getattr(profile, "dimension_scalings", {}))

		for unit in getattr(profile, "units_custom", []):
			p + unit

		p << getattr(profile, "parameters", {})

		p.cache(**getattr(profile, "parameters_cache", {}))

		p & getattr(profile, "parameters_units", {})

		p.set_units_context(getattr(profile, "units_context", None))

		return p

	def __rshift__(self, other):

		if not isinstance(other, str_types):
			raise errors.ParametersException("The right shift operator is used to save the parameters to a file. The operand must be a filename.")

		self.__save__(other)
		return self

	def __save__(self, filename):
		f = open(filename, 'w')

		# Export unit context
		if self.units_context is not None:
			f.write('units_context = %s' % self.units_context)

		# Export dimension scalings
		f.write("dimension_scalings = {\n")
		for dimension, scaling in self.__scalings.items():
			f.write("\t\"%s\": (%s,\"%s\"),\n" % (dimension, scaling.value, scaling.units))
		f.write("}\n\n")

		# Export custom units
		f.write("units_custom = {\n")
		for unit in self.__units_custom:
			f.write("%s,\n" % unit)
		f.write("}\n\n")

		# Export parameters
		f.write("parameters = {\n")
		for pam, value in self.__parameters.items():
			f.write("\t\"%s\": (%s,\"%s\"),\n" % (pam, value.value, value.units))
		f.write("}\n\n")

		# Export parameters_cache
		f.write("parameters_cache = {\n")
		for pam in self.__cache_funcs:
			f.write("\t\"%s\": True,\n" % (pam))
		f.write("}\n\n")

		# Export parameters_units
		f.write("parameters_units = {\n")
		for pam, units in self.__parameters_spec.items():
			if pam not in self.__parameters:
				f.write("\t\"%s\": \"%s\",\n" % (pam, units))
		f.write("}\n\n")

		f.close()


class Bounds(object):
	'''
	Bounds(param, units, bounds, error=True, clip=False, inclusive=True)

	:param param: Name of param to which bound applies.
	:type param: str
	:param units: Units of parameter.
	:type units: str or :class:`Units`
	:param bounds: The bound specification. See :func:`Parameters.bounds`.
	:type bounds: List of bounds.
	:param error: A boolean flag named error.
	:type error: bool
	:param clip: A boolean flag named clip.
	:type clip: bool
	:param inclusive: A boolean flag named inclusive.
	:type inclusive: bool

	The :class:`Bounds` object is used by a :class:`Parameters` instance to store
	the properties of a parameter bound. This is the object type returned by
	:func:`Parameters.bounds`. The above parameters can be extracted using attributes
	(if b is an instance of Bounds):

	>>> b.param
	>>> b.units
	>>> ...
	'''

	def __init__(self, param, units, bounds, error=True, clip=False, inclusive=True):
		self.param = param
		self.units = units
		self.bounds = bounds
		self.error = error
		self.clip = clip
		self.inclusive = inclusive



#!/usr/bin/env python
import sys
COLOURS = (
    'BLACK', 'RED', 'GREEN', 'YELLOW',
    'BLUE', 'MAGENTA', 'CYAN', 'WHITE'
)


def colour_text(text, colour_name='WHITE', bold=False):
    if colour_name in COLOURS:
        return '\033[{0};{1}m{2}\033[0m'.format(
            int(bold), COLOURS.index(colour_name) + 30, text)
    sys.stderr.write('ERROR: "{0}" is not a valid colour.\n'.format(colour_name))
    sys.stderr.write('VALID COLOURS: {0}.\n'.format(', '.join(COLOURS)))


# TESTS
if __name__ == '__main__':
    for bold in (False, True):
        for colour_name in COLOURS:
            print colour_text('Example of {0}'.format(colour_name), colour_name, bold)
    print
    # test error handling
    colour_text('TEST', 'SILVER')






# Root class for all exceptions generated by the Parameters module
class ParametersException(Exception):
	pass


# Parameter Errors
class ParameterValueError(ParametersException):
	pass


class ParameterOverSpecifiedError(ParametersException):
	pass


class ParameterNotInvertableError(ParametersException):
	pass


class ParameterInvalidError(ParametersException):
	pass


class ParameterRecursionError(ParametersException):
	pass


class ParameterNameWarning(UserWarning):
	pass


class ParameterInconsistentWarning(UserWarning):
	pass


# Bound Errors
class ParameterOutsideBoundsWarning(UserWarning):
	pass


class ParameterOutsideBoundsError(ParametersException):
	pass


class ParameterBoundsUncheckedWarning(UserWarning):
	pass


# Symbolic evaluation errors
class SymbolicEvaluationError(ParametersException):
	pass


class ExpressionOptimisationError(ParametersException):
	pass


# Quantity Errors
class QuantityValueError(ParametersException):
	pass


class QuantityCoercionError(ParametersException):
	pass


# Unit Errors
class UnitInvalidError(ParametersException):
	pass


class UnitConversionError(ParametersException):
	pass


# Scaling Errors
class ScalingValueError(ParametersException):
	pass


class ScalingUnitInvalidError(ParametersException):
	pass


class ScalingDimensionInvalidError(ParametersException):
	pass



from fractions import Fraction
import re, types, inspect

from . import errors
from .text import colour_text
from .utility.compat import UnicodeMixin, str_types


class Unit(UnicodeMixin):
	'''
	Unit (name,abbr=None,rel=1.0,prefixable=True,plural=None,dimensions={},base_unit=None)

	:class:`Unit` is the fundamental object which represents a single unit in units
	specification. For example, in "m*s", each of "m" and "s" would have their own
	associated :class:`Unit` object.

	:param name: A string or list of strings that are used as full names for this unit. If more than one is provided, the first is used as the default representation.
	:type name: str or list of str
	:param abbr: A string or list of strings that will be used to represent the abbreviated unit. If more than one is provided, the first is used as the default representation.
	:type abbr: str or list of str
	:param rel: The size of this unit compared to some fixed arbitrary basis.
	:type rel: float
	:param prefixable: Whether or not the unit can be prefixed (e.g. milli, micro, etc).
	:type prefixable: bool
	:param plural: When written in full, use this string as the plural unit. Note that when this is not provided, where implemented, an 's' is appended to the unit.
	:type plural: str
	:param dimensions: A dictionary specification of the dimensions of this unit.
	:type dimensions: dict
	:param base_unit: A Unit object corresponding to the base unit (if this unit is prefixed), or None otherwise.
	:type base_unit: Unit

	To instantiate a :class:`Unit` object, you can use something like:
	>>> u = Unit('metre',abbr='m',rel=1.0,dimensions={'length':1})

	Note that the keys for the dimensions can be whatsoever you please, and that
	you can invent your own dimensions. For example, you could create a unit with
	dimensions of "intelligence". The :class:`UnitDispenser` instance will pick up
	this dimension and handle unit conversions/etc, where that makes sense. The
	values of the dimensions dictionary indicate the power to which the dimension
	specified in the key should be taken.
	'''

	def __init__(self, name=None, abbr=None, rel=1.0, prefixable=True, plural=None, dimensions=None, base_unit=None):
		self.name = name
		self.abbr = abbr
		self.plural = plural

		self.rel = rel
		self.prefixable = prefixable

		self.dimensions = dimensions

		self.base_unit = base_unit if base_unit is not None else self

	@property
	def name(self):
		'''
		The default full name for this unit.

		To set the name(s) for this unit, you can use:

		>>> u.name = "meter"

		Or, if you want to provide multiple names as an alias.

		>>> u.name = ["meter","metre"]
		'''
		return self.__names[0]
	@name.setter
	def name(self, name):
		if name is None:
			raise ValueError("Unit objects require a valid name.")
		self.__names = name if isinstance(name, (tuple, list)) else (name,)
	@property
	def names(self):
		'''
		The full list or tuple of names specified for this unit.
		'''
		return self.__names

	@property
	def abbr(self):
		'''
		The default abbreviation for this unit.

		To set the name(s) for this unit, you can use:

		>>> u.abbr = "m"

		Or, if you want to provide multiple names as an alias.

		>>> u.abbr = ["m","<abbr>"]
		'''
		if self.__abbrs is None:
			return None
		return self.__abbrs[0]
	@abbr.setter
	def abbr(self, abbr):
		if abbr is None:
			self.__abbrs = None
		else:
			self.__abbrs = abbr if isinstance(abbr, (tuple, list)) else (abbr,)
	@property
	def abbrs(self):
		'''
		abbrs(self)

		:returns: The full list or tuple of abbreviations specified for this unit.
		'''
		return self.__abbrs

	@property
	def dimensions(self):
		'''
		The dictionary of dimensions specified for this unit.

		To set the dimensions for this unit, you can use:

		>>> u.dimensions = {'length':1}

		For more about the dimensions format, refer to comments in the general
		documentation for this class.
		'''
		return self.__dimensions
	@dimensions.setter
	def dimensions(self, dimensions):
		if dimensions is None:
			dimensions = {}
		if type(dimensions) != dict:
			raise ValueError("Invalid specications of unit dimensions.")
		self.__dimensions = dimensions
	def set_dimensions(self, **dimensions):
		'''
		set_dimensions(**dimensions)

		:param dimensions: Dictionary of dimensions.
		:type dimensions: dict

		:returns: A reference to the unit object.

		This is a shorthand convenience method that allows you to construct
		units in a slightly simpler form:

		>>> u = u = Unit('metre',abbr='m',rel=1.0).set_dimensions(length=1)
		'''
		self.__dimensions = dimensions
		return self

	def __repr__(self):
		return str(self)

	def __unicode__(self):
		return self.name


class UnitDispenser(UnicodeMixin):
	'''
	UnitDispenser()

	A :class:`UnitDispenser` instance is an object that manages a collection of
	:class:`Unit` objects.

	One can subclass the UnitDispenser object and implement the
	UnitsDispenser.[init_prefixies,init_units] methods to prepopulate the
	UnitDispenser object with units. This is done for the SI unit system with the
	:class:`SIUnitDispenser` class.

	Most of the functionality of the UnitDispenser class is clear from its methods.
	The functionality which is not clear from reading the documentation of the methods
	is introduced here.

	Creating a UnitDispenser instance:
		>>> ud = UnitDispenser()

	Generating Units objects:
		To generate a :class:`Units` object which represents a particular combination
		of individual :class:`Unit` objects, you can call this class with a string
		representing the units. For example:

		>>> ud('m*s')
		m*s

		This is shorthand for calling:

		>>> Units('m*s',dispenser=ud)
		m*s

		In addition to being shorter, the Units object is cached in the UnitDispenser
		object for future recall. See the documentation for :class:`Units` for more.

		For non-compound units, you can also use:

		>>> ud.km

		Which is shorthand for:

		>>> ud('km')
	'''

	def __init__(self):
		self._dimensions = {}
		self._units = {}
		self._prefixes = []

		self._contexts = {}
		self._context_current = False
		self._scalings = {}
		self._conversions = {}
		self.__convertable_units = []
		self.__conversions_cache = {}

		self.__cache = {}

		self.init_prefixes()
		self.init_units()

	############# SETUP ROUTINES ###########################################
	def init_prefixes(self):
		'''
		init_prefixes()

		A stub to allow subclasses to populate themselves.
		'''
		pass

	def init_units(self):
		'''
		init_units()

		A stub to allow subclasses to populate themselves.
		'''
		pass

	def add(self, unit, check=True):
		'''
		add(unit, check=True)

		:param unit: The unit to be added to the UnitDispenser.
		:type unit: Unit
		:param check: Whether to check that there is a basis unit for every dimension used by the :class:`Unit` object.
		:type check: bool

		:returns: A reference to self, for chaining.

		This method adds the supplied :class:`Unit` object to the dispenser's
		catalog of known units. If the unit object has a dimension of type never
		seen before by the dispenser, and the unit has no other dimensions, it is
		made the default "basis" for that dimension. If the unit has other dimensions
		as well, then a warning is printed stating that no basis element is known for that
		dimension.

		To add a unit (or units), you can use:

		>>> ud.add( Unit('metre',abbr='m',rel=1.0).set_dimensions(length=1,mass=-2) ).add( .... )

		You can also use the shorthand:

		>>> ud + Unit('metre',abbr='m',rel=1.0).set_dimensions(length=1,mass=-2) + .....

		.. warning:: When adding a unit, the dispenser will replace any existing unit by the same name; and that if Unit.prefixable is True, then it will add the unit as well as all possible prefixed versions of that unit.
		'''
		if not isinstance(unit, Unit):
			raise errors.UnitInvalidError("A Unit object is required for addition to a UnitDispenser. Was provided with: '%s'." % unit)

		for name in unit.names:
			self._units[name] = unit
		if unit.abbr != None:
			for abbr in unit.abbrs:
				self._units[abbr] = unit

		for dimension in unit.dimensions:
			if dimension not in self._dimensions or self._dimensions[dimension] is None:
				if unit.dimensions == {dimension: 1}:
					self.basis(**{dimension: unit})
				else:
					self._dimensions[dimension] = None
		if check:
			for dimension, basis_unit in self.basis().items():
				if basis_unit is None:
					print colour_text("WARNING: No basis unit specified for: %s." % dimension)

		if unit.prefixable:
			for prefix in self._prefixes:
				self.add(
					Unit(
						name=self.__generate_units(unit.names, prefix[0]),
						abbr=self.__generate_units(unit.abbrs, prefix[1]),
						plural=self.__generate_units(unit.plural, prefix[0]),
						rel=unit.rel * prefix[2],
						prefixable=False,
						base_unit=unit
					).set_dimensions(**unit.dimensions),
					check=False)

		return self

	def add_context(self, *name, **params):
		if params == {} and len(name) == 2:
			params = name[1]
			name = name[0]
		else:
			assert(len(name) == 1)
			name = name[0]
		self._contexts[name] = params

	def set_context(self, *name, **params):
		assert(len(name) == 1)
		name = name[0]

		if name not in self._contexts:
			raise ValueError("Non-existent context %s" % name)

		ps = self._contexts[name].copy()

		for p in ps.keys():
			if p in params:
				ps[p] = params[p]

		self._context_current = (name, ps)
		self.__cache = {}

	@property
	def context(self):
		if self._context_current is False:
			return None
		else:
			return self._context_current

	def add_scaling(self, dim_from, dim_to, scaling, context=None):
		'''
		add_scaling(dim_from, dim_to, scaling, context=None)

		:param dim_from: A dictionary of dimensions from which scaling will be provided to `dim_to`.
		:type dim_from: dict
		:param dim_to: A dictionary of dimensions to which scaling will be provided from `dim_from`.
		:type dim_to: dict
		:param scaling: The scaling to be acquired.
		:type scaling: float
		:param context: The context to which this scaling belongs.
		:type context: string

		This method adds a scaling between units of different dimensions that
		can be used to simplify computations in different physical contexts.
		For example, in many condensed matter physics, energies are often
		discussed as frequencies (and vice versa).

		Using:

		>>> ud.add_scaling( ud.joule.dimensions, ud.hertz.dimensions, 1./2/hbar/pi, context='condensed_matter' )

		Will allow you to do:

		>>> SIQuantity( 2, 'J' )('GHz', context='condensed_matter')
		'''
		assert(type(dim_from) == dict)
		assert(type(dim_to) == dict)
		if context not in self._contexts:
			self._contexts[context] = {}
		if context not in self._scalings:
			self._scalings[context] = []
		self._scalings[context].append( (dim_from, dim_to, scaling) )

	def is_scalable(self, dim_from, dim_to, context=False):
		'''
		is_scalable(dim_from, dim_to, context=False)

		:param dim_from: A dictionary of dimensions from which scaling to `dim_to` will be tested.
		:type dim_from: dict
		:param dim_to: A dictionary of dimensions to which scaling from `dim_from` will be tested.
		:type dim_to: dict
		:param context: The context in which to test this scaling.
		:type context: str

		This method returns `True` if there is a special scaling specified
		between the dimensions provided.
		'''
		try:
			self.scale(dim_from, dim_to, context=context)
			return True
		except:
			return False

	def scale(self, dim_from, dim_to, context=False):
		'''
		scale(dim_from, dim_to, context=False)

		:param dim_from: A dictionary of dimensions from which to scale.
		:type dim_from: dict
		:param dim_to: A dictionary of dimensions to which to scale.
		:type dim_to: dict
		:param context: The context in which to provide this scaling, or False.
		:type context: str

		This method returns a float corresponding to a scaling from the
		dimensions provided, as specified using `UnitDispenser.add_scaling`.
		'''

		if context is False:
			if self._context_current is not False:
				context = self._context_current[0]
			else:
				context = None

		if context in self._scalings:
			for scaling in self._scalings[context]:
				if dim_from == scaling[0] and dim_to == scaling[1]:
					return self.__eval_context_function(scaling[2], context)
				elif dim_from == scaling[1] and dim_to == scaling[0]:
					return 1./self.__eval_context_function(scaling[2], context)

		if context is not None:
			return self.scale(dim_from, dim_to, context=None)

		raise ValueError("No scaling between dimensions %s and %s are possible." % (dim_from, dim_to))

	def __eval_context_function(self, f, context, args=(), kwargs={}):
		if not type(f) is types.FunctionType:
			return f

		if f.__code__ is None:
			raise ValueError("Error during evaluation of function, due to its being non-introspectable. Most likely, this function is a cython function.")

		args = list(args)
		arg_names = f.__code__.co_varnames[len(args):f.__code__.co_argcount]

		for arg in arg_names:
			if arg in kwargs:
				args.append(kwargs[arg])
			elif self._context_current is not False and context == self._context_current[0] and arg in self._context_current[1]:
				args.append(self._context_current[1][arg])
			elif context in self._contexts and arg in self._contexts[context]:
				args.append(self._contexts[context][arg])
			else:
				raise ValueError("Value for `%s` not available in context." % arg)

		return f(*args)

	def add_conversion_map(self, unit_from, unit_to, mapping, absolute=False, context=None):
		'''
		add_conversion_map(unit_from, unit_to, mapping, absolute=True, context=None)

		:param unit_from: The units from which to convert.
		:type dim_from: str or Units
		:param unit_to: The units to which to convert.
		:type dim_to: str or Units
		:param mapping: A callable that takes a single value in units `unit_from`
			and returns a value with units `unit_to`.
		:type mapping: callable
		:param absolute: Whether the map provided is for units with an absolute scale
			or a relative one (i.e. fahrenheit -> celcius depends on whether
			stored values are absolute or delta values).
		:type absolute: bool
		:param context: The context in which to provide this conversion.
		:type context: str

		This method adds a potentially non-linear mapping between units that
		can be used to simplify computations in different physical contexts.
		For example, it may be useful to use store data in 'decibels', and later
		convert this to a linear value. This can be handled automatically.

		Using:

		>>> ud.add_scaling( 'dB', '', lambda v: 10**(v/10.), absolute=True, context=None )

		Will allow you to do:

		>>> SIQuantity( 1, 'dB' )('')

		'''
		# TODO: Add checks
		if context not in self._contexts:
			self._contexts[context] = {}
		if context not in self._conversions:
			self._conversions[context] = []
		self._conversions[context].append( ( self(unit_from), self(unit_to), mapping, absolute) )
		self.__convertable_units.append(self(unit_from))

	def has_conversion_map(self, unit_from, unit_to, absolute=False, context=False):
		'''
		has_conversion_map(unit_from, unit_to, absolute=True, context=None)

		:param unit_from: The units from which to convert.
		:type dim_from: str or Units
		:param unit_to: The units to which to convert.
		:type dim_to: str or Units
		:param absolute: Whether the map found is an an absolute conversion
			or a relative one (i.e. fahrenheit -> celcius depends on whether
			stored values are absolute or delta values).
		:type absolute: bool
		:param context: The context in which to provide this conversion.
		:type context: str

		This method returns `True` if there exists a conversion between the provided
		units, as specified using `UnitDispenser.add_conversion_map`.
		'''
		try:
			self.conversion_map(unit_from, unit_to, absolute=True, context=context)
			return True
		except:
			return False

	def conversion_map(self, unit_from, unit_to, absolute=False, context=False):
		'''
		conversion_map(unit_from, unit_to, absolute=True, context=None)

		:param unit_from: The units from which to convert.
		:type dim_from: str or Units
		:param unit_to: The units to which to convert.
		:type dim_to: str or Units
		:param absolute: Whether the map found should be for an absolute conversion
			or a relative one (i.e. fahrenheit -> celcius depends on whether
			stored values are absolute or delta values).
		:type absolute: bool
		:param context: The context in which to provide this conversion.
		:type context: str

		This method returns a function that when called upon the value associated
		with a quantity of units `unit_from` will return the appropriate value
		for a new quantity with units `unit_to`. This method only returns
		such a function for definitions provided to `UnitDispenser.add_conversion_map`.
		'''
		unit_from, unit_to = self(unit_from), self(unit_to)
		if unit_from not in self.__convertable_units:
			raise ValueError("No mapping known between %s and %s" % (unit_from, unit_to))

		if context is False:
			if self._context_current is not False:
				context = self._context_current[0]
			else:
				context = None

		check = False
		try:
			c = self.__conversions_cache[context][unit_from][unit_to][absolute]
			if c is None:
				check=False
			else:
				return c
		except KeyError:
			check = True

		if check and context in self._conversions:
			unit_from, unit_to = self(unit_from), self(unit_to)
			m = lambda u: u.base_unit.name

			c = None
			for conversion in self._conversions[context]:
				if conversion[-1] != absolute:
					continue
				# Try to convert current unit to stored unit
				try:
					assert(sorted(map(m, conversion[0].units.keys())) == sorted(map(m, unit_from.units.keys())))
					assert(sorted(map(m, conversion[1].units.keys())) == sorted(map(m,unit_to.units.keys())))
					pre_scaling = unit_from.scale(conversion[0])
					post_scaling = unit_to.scale(conversion[1])

					c = lambda v: self.__eval_context_function(conversion[2], context, args=[pre_scaling*v])/post_scaling

					break

				except:
					pass

			if context not in self.__conversions_cache:
				self.__conversions_cache[context] = {}
			if unit_from not in self.__conversions_cache[context]:
				self.__conversions_cache[context][unit_from] = {}
			if unit_to not in self.__conversions_cache[context][unit_from]:
				self.__conversions_cache[context][unit_from][unit_to] = {}
			self.__conversions_cache[context][unit_from][unit_to][absolute] = c

			if c is not None:
				return c

		if context is not None:
			return self.conversion_map(unit_from, unit_to, context=None)

		raise ValueError("No mapping known between %s and %s" % (unit_from, unit_to))

	def __generate_units(self, names, prefixes):
		if names is None or prefixes is None:
			return None

		if not type(names) in (list,tuple):
			names = (names,)
		if not type(prefixes) in (list,tuple):
			prefixes = (prefixes,)

		units = []
		for name in names:
			for prefix in prefixes:
				units.append("%s%s" % (prefix, name))
		return units

	def __add__(self, unit):
		self.add(unit)
		return self

	def list(self):
		'''
		list()

		:returns: A list of strings representing the units recognised by the UnitDispenser.

		For example:

		>>> ud.list()
		['m','cm','mm',...]

		Note that no particular order is guaranteed for this list.
		'''
		return self._units.keys()

	def has(self, identifier):
		'''
		has(identifier)

		:param identifier: A string representation of the unit of interest.
		:type identifier: str

		:returns: :python:`True` if the :class:`UnitDispenser` object recognises the string representation; and :python:`False` otherwise.

		For example:

		>>> ud.has('metre')
		'''

	def get(self, unit):
		'''
		get(unit)

		:param unit: A string representation of the unit of interest.
		:type unit: str

		:returns: :class:`Unit` object associated with a the string representation.
		:raises: :class:`UnitInvalidError` if no unit can be found that matches.
		'''
		if isinstance(unit, str_types):
			try:
				return self._units[unit]
			except:
				raise errors.UnitInvalidError("Unknown unit: '%s'." % unit)
		elif isinstance(unit, Unit):
			return unit
		raise errors.UnitInvalidError("Could not find Unit object for '%s'." % unit)

	@property
	def dimensions(self):
		'''
		A list of known dimensions. For example:

		>>> ud.dimensions = ['length', 'mass', 'time']
		'''
		return self._dimensions.keys()

	def basis(self, **kwargs):
		'''
		basis(**kwargs)

		:param kwargs: A dictionary of :class:`Unit` or :class:`str` objects with dimension names as keys.
		:type kwargs: dict

		This method is used to both extract and set the "basis" units for this dispenser
		instance. By default, the basis is the set of Unit objects first set with dimensions
		of {<dimension> : 1}. The only role the basis is plays is to set a standard
		canonical basis of units in which to express physical quantities.

		To examine the current basis, use:

		>>> ud.basis()

		To update or modify the basis, use:

		>>> ud.basis(length='cm', time='ns')

		For more about how this useful, see the documentation for :python:`Quantity.basis`.
		'''
		if not kwargs:
			return self._dimensions

		for key, val in kwargs.items():
			unit = self.get(val)
			if unit.dimensions == {key: 1}:
				self._dimensions[key] = unit
			else:
				print "Invalid unit (%s) for dimension (%s)" % (unit, key)

	############# UNITS GENERATION #########################################

	def __call__(self, units):
		'''
		This is a shortcut for: Units(units,dispenser=self); which also allows
		for caching.
		'''
		if type(units) is str:
			if units in self.__cache:
				return self.__cache[units]
			self.__cache[units] = Units(units, dispenser=self)
			return self.__cache[units]
		return Units(units, dispenser=self)

	def __getattr__(self, name):
		if name[:2] == "__" or name[:14] == "_UnitDispenser":
			raise AttributeError
		return self(name)


class Units(UnicodeMixin):
	'''
	Units(units=None,dispenser=None)

	The :class:`Units` object is the highest level class in python-parameters which
	deals only with units; and is the class used directly by :class:`Quantity`. It
	handles drawing appropriate units from a :class:`UnitDispenser`, and performing
	:class:`Unit` arithmetic. This arithmetic is mostly useful when used behind
	the scenes by a :class:`Quantity` instance.

	:param units: A representation of the units in some form. See below for details.
	:param dispenser: A reference to the UnitDispenser from which to draw units for this object.
	:type dispenser: UnitDispenser

	Representation of units:
		The unit representation passed to the :class:`Units` constructor can be:
			- A Units object (in which case a copy is returned)
			- A Unit object (in which case it is upgraded to a :class:`Units` object)
			- A string representing a units object
			- A dictionary of Unit-power relationships

		A valid string representation is a string which consists of a series of
		unit tokens (described below) separated by either a "* (for multiplication)
		or a "/" (for division). Each unit token consists of a unit string
		representation recognisable by a dispenser (for example: "ms" or "millisecond")
		followed by an optional power; which is indicated by a caret "^" and a floating
		point number (including integers). Importantly fractions are not supported in this
		specification.

		For example, here are some valid units string representations:
			- "kg*m*s^-2"
			- "ms/nm*kg^2"
			- "/nm"

		A valid dictionary of unit-power relationships is a mapping from :class:`Unit`
		object keys (or their string representation) to a numeric power. For example:

		>>> Units(units={"ms":1,"kg":0.5}, dispenser=ud)

	Units arithemetic:
		Units objects support the following arithmetic operations:
		- Multiplication
		- Division
		- Arbitrary powers

		For example:

		>>> ud = SIUnitDispenser()
		>>> u1 = Units('J', dispenser=ud)
		>>> u2 = Units('nm*s', dispenser=ud)
		>>> u1*ud
		J*s*nm
		>>> u1/ud
		J/s/nm
		>>> u2**2
		s^2*nm^2

		If you multiply or divide a unit object by an object other than a `Units`
		instance, a `Quantity` object will be returned. For example:

		>>> 3 * ud.nm
		3 nm

		>>> ud.nm**2 / 2
		0.5 nm^2

	Units equality:
		Units objects can also recognise when they are equal to other Units objects.

		For example (with u1 and u2 as defined above):

		>>> u1 == u2
		False
		>>> u1 != u2
		True


	There are a few useful methods though.

	To determine the numerical scaling factor between two units, you can use:
	>>> units.scale(other_units)
	This will raise an exception if the other units have different dimensions.

	The numerical scaling factor of this unit relative to the unit basis of
	the unit dispenser is given by:
	>>> units.rel

	The dimensions of the units object is given by:
	>>> units.dimensions

	The unit equivalent unit in the basis of the UnitDispenser is given by:
	>>> units.basis

	And you can see the dependence of Units on the underlying Unit objects
	directly by using:
	>>> units.units
	'''

	def __init__(self, units=None, dispenser=None):
		self.__hash = hash(str(units))
		self.__dispenser = dispenser
		self.__units = self.__process_units(units)

	def __get_unit(self, unit):
		return self.__dispenser.get(unit)

	def __process_units(self, units):

		if units is None:
			return {}

		elif isinstance(units, Units):
			return units.units.copy()

		elif isinstance(units, Unit):
			return {units: 1}

		elif isinstance(units, dict):
			units = units.copy()
			for unit in units.keys():
				if type(unit) != Unit:
					units[self.__get_unit(unit)] = Fraction(units.pop(unit))
			return units

		elif isinstance(units, str_types):
			d = {}

			if units == "units":
				return d

			for match in re.finditer("([*/])?([^*/\^0-9]+)(?:\^(\-?[0-9\.]+))?", units.replace(" ", "")):
				groups = match.groups()
				mult = -1 if groups[0] == "/" else 1
				power = mult * (Fraction(groups[2]) if groups[2] is not None else 1)

				unit = self.__get_unit(groups[1])
				d[unit] = d.get(unit, 0) + power

			return d

		raise errors.UnitInvalidError("Unrecognised unit description %s" % units)

	def __repr__(self):
		return str(self)

	def __unicode__(self):
		output = []

		items = sorted(self.__units.items(), key=str)

		for unit, power in items:
			if power > 0:
				if power != 1:
					output.append("%s^%s" % (unit.abbr, power))
				else:
					output.append(unit.abbr)
		output = "*".join(output)

		for unit, power in items:
			if power < 0:
				if power != -1:
					output += "/%s^%s" % (unit.abbr, abs(power))
				else:
					output += "/%s" % unit.abbr

		return output

	def scale(self, other, context=False, value=None):
		'''
		scale(other)

		:param other: Unit with which to compare.
		:type other: Units

		:returns: A float such that when multiplying the value of a physical quantity with this objects units it would return the value of the same physical quantity represented in 'other's units.
		:raises: UnitConversionError if "other" does not have the same dimensions as this object.

		This method returns the scaling between the units of this object and that of another.
		This method is used by the :class:`Quantity` object in order to provide unit
		conversion.

		E.g.: <km>.scale('m') -> 1000
		'''
		try:
			return self.__scale_cache[other]
		except:
			if getattr(self, '__scale_cache', None) is None:
				self.__scale_cache = {}

			if isinstance(other, str_types):
				other = self.__dispenser(other)

			dims = self.dimensions
			dims_other = other.dimensions

			scale = None
			try:
				scale = self.__dispenser.scale(dim_from=dims, dim_to=dims_other, context=context)
			except ValueError as e:
				pass

			# If the union of the sets of dimensions is less than the maximum size of the dimensions; then the units are the same.
			if scale is None:
				if len(set(dims.items()) & set(dims_other.items())) < max(len(dims), len(dims_other)):
					raise errors.UnitConversionError("Invalid conversion. Units '%s' and '%s' do not match. %s" % (self, other, set(dims.items()) & set(dims_other.items())))
				self.__scale_cache[other] = self.rel / other.rel
			else:
				return self.rel / other.rel * scale # Don't cache if scaling was applied.
			return self.__scale_cache[other]

	@property
	def dimensions(self):
		'''
		A dictionary representing the dimensions of the units described by this object.

		>>> u.dimensions
		{'length': 1, 'time': 1}
		'''

		dimensions = {}
		for unit, power in self.__units.items():
			for key, order in unit.dimensions.items():
				dimensions[key] = dimensions.get(key, 0) + power * order
		for key, value in list(dimensions.items()):
			if value == 0:
				del dimensions[key]

		return dimensions

	@property
	def rel(self):
		'''
		The relative size of this unit (compared to other units from the same unit dispenser, and with respect to an arbitrary fixed basis).

		>>> u.rel = 1.2
		'''

		rel = 1.
		for unit, power in self.__units.items():
			rel *= unit.rel ** power
		return rel

	def basis(self):
		'''
		basis()

		:returns: A new :class:`Units` object representing the units in the :class:`UnitDispenser` basis that correspond to the same dimensions as this object.

		For example:

		>>> Units('g*ns',dispenser=SIUnitDispenser()).basis()
		'kg*s'
		'''
		dimensionString = ""
		dimensionMap = self.__dispenser.basis()
		dimensions = self.dimensions

		for dimension in self.dimensions:
			if dimensions[dimension] != 0:
				dimensionString += "*%s^%f" % (dimensionMap[dimension].abbr if dimensionMap[dimension].abbr is not None else dimensionMap[dimension], float(dimensions[dimension]))

		return Units(dimensionString[1:], dispenser=self.__dispenser)

	@property
	def units(self):
		'''
		A dictionary representation of the units of this object.

		>>> u.units
		{ms: 1, kg: -1}
		'''
		return self.__units.copy()

	########### UNIT OPERATIONS ############################################

	def __new(self, units):
		return Units(units, self.__dispenser)

	def __mul_units(self, target, additive):
		for unit, power in additive.items():
			target[unit] = target.get(unit, 0) + power

			if target.get(unit, 0) == 0:
				del target[unit]
		return target

	def __div_units(self, target, additive):
		for unit, power in additive.items():
			target[unit] = target.get(unit, 0) - power

			if target.get(unit, 0) == 0:
				del target[unit]
		return target

	def __mul__(self, other):
		if not isinstance(other,Units):
			from .quantities import Quantity
			return Quantity(other, self, dispenser=self.__dispenser)
		return self.__new(self.__mul_units(self.units, other.units))

	def __rmul__(self,other):
		from .quantities import Quantity
		return Quantity(other, self, dispenser=self.__dispenser)

	def __div__(self, other):
		if not isinstance(other,Units):
			from .quantities import Quantity
			return Quantity(1./other, self, dispenser=self.__dispenser)
		return self.__new(self.__div_units(self.units, other.units))

	def __rdiv__(self,other):
		if other == 1:
			return self.__pow__(-1)
		from .quantities import Quantity
		return Quantity(other, 1/self, dispenser=self.__dispenser)

	def __truediv__(self, other):
		return self.__div__(other)

	def __rtruediv__(self, other):
		return self.__rdiv__(other)

	def __pow__(self, other):
		new_units = self.units
		for unit in new_units:
			new_units[unit] *= other
		return self.__new(new_units)

	def __eq__(self, other):
		if str(self) == str(other):
			return True
		return False

	def __hash__(self):
		return self.__hash



import sys
import threading
import resource
import datetime
import types

import numpy as np


class RangesIterator(object):
	'''
	RangesIterator(parameters, ranges, params={}, masks=None, function=None, function_args=(), function_kwargs={}, nprocs=None, distributed=False, ranges_eval=None, progress=True)

	:class:`RangesIterator` is a python iterable object, which allows one to easily
	iterate over a potentially multidimensional space of parameters. It also has
	inbuilt multithreading 	support which is used when a function is supplied in order
	to fully take advantage of the available computational resources.

	:param parameters: A reference to a Parameters instance.
	:type parameters: Parameters
	:param ranges: The ranges to iterate over.
	:type ranges: dict or list of dict
	:param params: A background context of parameters upon which the ranges should be superimposed.
	:type params: dict
	:param masks: A list of boolean functions which indicate when a particular index should be computed.
	:type masks: list of callable objects
	:param function: An (optional) function to call for every resulting parameter configuration.
	:type function: callable
	:param function_args: An (optional) tuple/list of args to pass to the above function at every call.
	:type function_args: tuple or list
	:param function_kwargs: An (optional) dictionary of kwargs to pass to the above function at every call.
	:type function_kwargs: dict
	:param nprocs: The number of processes to spawn at any one time (for multithreading support).
	:type nprocs: None or int
	:param distributed: `False` or `None` if distributed computing using dispynode servers is NOT to be used,
		and `True` or a dictionary of dispy.JobCluster parameters otherwise.
	:type distributed: NoneType, bool or dict
	:param ranges_eval: An (optional) previously computed ranges_eval to use in this enumeration.
	:type ranges_eval: numpy.ndarray
	:param progress: `True` if progress should be shown, and `False` otherwise. This can also
		be a callable object taking arguments `total`, `completed` and `start_time`, which
		are the total number of indices to compute, the number completed computations,
		and the start time computed using `datetime.datetime.now()`.
	:type progress: bool or callable

	Constructing a RangesIterator instance:
		In its simplest form, initialising a :class:`RangesIterator` looks like:

		>>> RangesIterator(p, ranges)

		Where *p* is a Parameters instance, and *ranges* is a valid specification
		of the ranges to be iterated over.

	Valid ranges specification:
		:python:`ranges` can be any valid range specification (as defined in
		:func:`Parameters.range`), or any list of valid range specifications.
		If a list of specifications is provided, then the cartesian product of
		range specifications is taken. For example:

		>>> ranges = [{'x':(0,10,11)},{'y':(0,10,11)}]

		If passed to RangesIterator, this will lead to iteration over the integer
		values of x in [0,10], and then for each x, to iterate over the integer values of
		y in [0,10]; forming the cartesian product [0,10]x[0,10].

		Remember that multiple parameters can be set at once, so that the following
		is also valid:
		>>> ranges = [{'x':(0,10,11),'k':(1,2,11)},{'y':(0,10,11)}]

	Iterating over a RangesIterator instance:
		To iterate over the possible parameter configurations, you use the regular
		iteration sytax:

		>>> for result in iterator:
				# Do something here

		In each iteration of the above loop, result will be a two-tuple of the indices of
		the current iteration in the cartesian product and one of the following:

		- If :func:`function` is not specified, then a dictionary of the parameters (including, except where overwritten, those in :python:`params`) corresponding to the cartesian indices.
		- If :func:`function` is specified, the value of the function evaluated in the parameter context, and with the kwargs arguments in :python:`function_kwargs`.

	Specifying a function to be evaluated:
		If specified, the :func:`function` value must be a callable type, such as a *function* or
		*instancemethod* (or any object implementing :func:`__call__`). The function must take at least the keyword value of
		:python:`params`. The following are examples of valid specifications:

		>>> def test(params):
				pass
		>>> RangesIterator(..., function=test, ...)

		>>> class Test(object):
				def test(self,params):
					pass
		>>> RangesIterator(..., function=Test().test, ...)

		>>> RangesIterator(..., function=lambda params: None, ...)

		The function can return any type, which will then be returned to you via
		the iteration process, as specified above.

		If :python:`function_kwargs` is specified, then the function provided
		will be evaluated with the additional kwargs present in that dictionary.

	Multithreading:
		By default, if :python:`function` is provided, :python:`RangesIterator`
		will spawn up to *N* parallel subprocesses to evaluate the function in different
		parameter contexts (where *N* is detailed below). If this is undesired, because your function is
		not threadsafe or for other reasons, you can specify :python:`nprocs` as
		a *0* or a *1*; in which case multithreading is disabled.

		Otherwise, :python:`nprocs` is to be the number of processes to use (with a default
		value of the number of processors of the machine). If specified as a negative
		number, then the iteration process will use that many fewer than the total number
		of processors on your machine.

	Distributed Computing:
		It is possible to have `RangesIterator` distribute tasks to any available dispynode
		servers. To enable this (which takes precedence over the above multithreading), simply
		set `distributed` to `True` or a dictionary of arguments to pass on to `dispy.JobCluster`.

	Masking:
		If you do not want the parameters or evaluated function at all possible
		cartesian products of the input ranges, then it is possible to use
		boolean masking functions. This is useful, for example, when wanting continue
		a previously started sweep.

		Masks should be callable objects with a signature of:
		:code:`<mask_name>(indices, ranges=None, params={})`

		For example:

		>>> def deny_mask(indices, ranges=None, params={}):
				return False # Deny all indices.

		At runtime, the current indices in question are passed as indices, along
		with the range specifications and current parameter context.
	'''

	def __init__(self, parameters, ranges, params={}, masks=None, function=None, function_args=(), function_kwargs={}, nprocs=None, distributed=False, ranges_eval=None, progress=True):
		self.p = parameters
		self.ranges = ranges
		self.function = function
		self.function_args = function_args
		self.function_kwargs = function_kwargs
		self.params = params
		self.masks = masks
		self.nprocs = nprocs
		self.distributed = distributed
		self.ranges_eval = ranges_eval
		self.progress = progress

		if current_process().name != "MainProcess" or threading.current_thread().name != "MainThread":
			self.nprocs = 1
			self.distributed = False

	@property
	def p(self):
		'''
		A reference to a Parameters instance.

		You can update this reference using:

		>>> iterator.p = <Parameters Instance>
		'''
		return self.__p
	@p.setter
	def p(self, parameters):
		self.__p = parameters

	@property
	def ranges(self):
		'''
		A reference to the list of ranges stored by the iterator.

		You can change the ranges used by the iterator using:

		>>> iterator.ranges = <Ranges specification>

		Note that this will reset :func:`ranges_eval`.
		'''
		return self.__ranges
	@ranges.setter
	def ranges(self, ranges):
		if isinstance(ranges, dict):
			ranges = [ranges]
		self.__ranges = ranges
		self.__ranges_eval = None

	@property
	def function(self):
		'''
		A reference to the callable to be called at each iteration, which is most likely a function or method (or None).

		You can change the callable using:

		>>> iterator.function = <callable>
		'''
		return self.__function
	@function.setter
	def function(self, function):
		if not isinstance(function, (type(None), types.FunctionType, types.MethodType)):
			raise ValueError("`function` must be a function or a method.")
		self.__function = function

	@property
	def function_kwargs(self):
		'''
		A reference to the dictionary of kwargs to passed to the callable at each iteration.

		You can change the kwargs using:

		>>> iterator.function_kwargs = <kwargs dict>
		'''
		return self.__function_kwargs
	@function_kwargs.setter
	def function_kwargs(self, function_kwargs):
		if not isinstance(function_kwargs, dict):
			raise ValueError("`function_kwargs` must be a dictionary of keyword arguments.")
		self.__function_kwargs = function_kwargs

	@property
	def params(self):
		'''
		A reference to the parameter context used by the iterator.

		You can change the parameter context using:

		>>> iterator.params = <parameters dictionary>

		Note that this will reset :func:`ranges_eval`.
		'''
		return self.__params
	@params.setter
	def params(self, params):
		if not isinstance(params, dict):
			raise ValueError("`params` must be a dictionary of parameter values (with parameters as keys).")
		self.__params = params
		self.__ranges_eval = None

	@property
	def masks(self):
		'''
		The list of mask callables used to filter which indices in the cartesian
		product of ranges are to be considered.

		You can change the masks using:

		>>> iterator.masks = <list of masks>
		'''
		return self.__masks
	@masks.setter
	def masks(self, masks):
		self.__masks = masks

	@property
	def nprocs(self):
		'''
		The number of processes to use (if positive), the number of CPUs to leave free
		(if negative), or None to default to using all CPUs.

		You can change nprocs using:

		>>> iterator.nprocs = <integer or None>
		'''
		return self.__nprocs
	@nprocs.setter
	def nprocs(self, nprocs):
		self.__nprocs = nprocs

	@property
	def ranges_eval(self):
		'''
		A previously evaluated ranges_eval from ranges_expand which can used to
		shortcut the range evaluation phase, and to ensure consistency when the
		range sampler is stochastic. If not specified when constructing the iterator,
		it will be generated when access it via this property, and then cached for later
		use.

		You can override ranges_eval using:

		>>> iterator.ranges_eval = <Previously generated ranges_eval>
		'''
		if self.__ranges_eval is None:
			self.__ranges_eval = self.ranges_expand()[0]
		return self.__ranges_eval
	@ranges_eval.setter
	def ranges_eval(self, ranges_eval):
		self.__ranges_eval = ranges_eval

	def ranges_expand(self):
		'''
		ranges_expand()

		:returns: A two-tuple of a structured numpy.ndarray with keys that are the parameters being iterated over and values being their current non-dimensional value, and the list of indices to consider as filtered by masks.

		For example:

		>>> iterator = RangesIterator(parameters=p, ranges=[{'x':(0,1,2)},{'y':(3,4,2)}])
		>>> iterator.ranges_expand()
		(array([[(0.0, 3.0), (0.0, 4.0)],
		        [(1.0, 3.0), (1.0, 4.0)]],
		       dtype=[('x', '<f8'), ('y', '<f8')]), [(0, 0), (0, 1), (1, 0), (1, 1)])
		'''
		return self.__ranges_expand(masks=self.masks, params=self.params.copy(), ranges_eval=self.__ranges_eval)

	@property
	def progress(self):
		'''
		A boolean indicating whether progress should be shown, or a callable object
		which takes arguments:
			- `total`: The total number of computations to be performed.
			- `completed`: The number of computations completed.
			- `start_time`: When the computation started (as a `datetime.datetime` object).

		You can change progress using:

		>>> iterator.progress = <bool or callable>
		'''
		return self.__progress
	@progress.setter
	def progress(self, progress):
		self.__progress = progress

	def __ranges_expand(self, level=0, iteration=tuple(), masks=None, indices=None, params=None, ranges_eval=None):
		'''
		This method generates a list of different parameter configurations
		'''
		if indices is None:
			indices = []
		if params is None:
			params = {}

		pam_ranges = self.ranges[level]

		# # Interpret ranges
		pam_values = {}
		count = None

		for param, pam_range in pam_ranges.items():
			if ranges_eval is not None and ranges_eval.ndim == len(self.ranges) and param in ranges_eval.dtype.fields.keys() and not np.any(np.isnan(ranges_eval[param])):  # If values already exist in ranges_eval, reuse them
				pam_values[param] = ranges_eval[param][iteration + (slice(None),) + tuple([0] * (ranges_eval.ndim - len(iteration) - 1))]

		tparams = params.copy()
		tparams.update(pam_ranges)
		tparams.update(pam_values)

		pam_values = self.p.range(list(pam_ranges.keys()), **tparams)

		c = len(pam_values[param])
		count = c if count is None else count
		if c != count:
			raise ValueError("Parameter ranges for %s are not consistent in count: %s" % (param, pam_ranges))

		if ranges_eval is None or ranges_eval.ndim < level + 1:
			ranges_eval = self.__extend_ranges(ranges_eval, list(pam_ranges.keys()), count)

		for i in range(count):
			current_iteration = iteration + (i,)

			# Generate slice corresponding all the components of range_eval this iteration affects
			s = current_iteration + tuple([slice(None)] * (ranges_eval.ndim - len(current_iteration)))

			# Update parameters
			for param, pam_value in pam_values.items():
				ranges_eval[param][s] = pam_value[i]
				if np.isnan(pam_value[i]):
					raise ValueError("Bad number for parameter %s @ indices %s" % (param, str(current_iteration)))

				params[param] = pam_value[i]

			if level < len(self.ranges) - 1:
				# Recurse problem
				ranges_eval, _ = self.__ranges_expand(level=level + 1, iteration=current_iteration, indices=indices, params=params, masks=masks, ranges_eval=ranges_eval)
			else:
				if masks is not None and isinstance(masks, list):
					if not any([mask(indices=current_iteration, ranges=self.ranges, params=params) for mask in masks]):
						continue

				indices.append(current_iteration)

		return ranges_eval, indices

	def __extend_ranges(self, ranges_eval, labels, size):
		dtype_delta = [(label, float) for label in labels]
		if ranges_eval is None:
			ranges_eval = np.zeros(size, dtype=dtype_delta)
			ranges_eval.fill(np.nan)
		else:
			final_shape = ranges_eval.shape + (size,)
			ranges_eval = np.array(np.repeat(ranges_eval, size).reshape(final_shape), dtype=ranges_eval.dtype.descr + dtype_delta)
		for label in labels:
			ranges_eval[label].fill(np.nan)
		return ranges_eval

	def __index_to_dict(self, index, ranges_eval):
		params = {}
		vs = ranges_eval[index]
		names = ranges_eval.dtype.names
		for i, param in enumerate(names):
			params[param] = vs[i]
		return params

	def __get_params_for_index(self, index, ranges_eval):
		params = self.params.copy()
		params.update(self.__index_to_dict(index, ranges_eval))
		return params

	def __iter__(self):
		ranges_eval, indices = self.ranges_expand()

		start_time = datetime.datetime.now()
		if self.distributed not in (None, False) and  self.function is not None:
			try:
				from .utility.symmetric import DistributedParallelMap
			except:
				raise RuntimeError("The `dispy` module is required for distributed iteration.")

			cluster_kwargs = {} if self.distributed is True else self.distributed
			dpm = DistributedParallelMap(self.function, progress=self.progress, **cluster_kwargs)

			for res in dpm.iterate([(i, self.function_args, {'params':self.__get_params_for_index(i, ranges_eval)}) for i in indices], count_offset=0, count_total=len(indices), start_time=start_time, base_kwargs=self.function_kwargs):
				yield res

		elif self.nprocs not in [0, 1] and self.function is not None:
			from .utility.symmetric import AsyncParallelMap
			apm = AsyncParallelMap(self.function, progress=self.progress, nprocs=self.nprocs, spawnonce=True)

			for res in apm.iterate([(i, self.function_args, {'params':self.__get_params_for_index(i, ranges_eval)}) for i in indices], count_offset=0, count_total=len(indices), start_time=start_time, base_kwargs=self.function_kwargs):
				yield res
		else:
			for i, index in enumerate(indices):
				if self.function is None:
					yield (index, self.__index_to_dict(index, ranges_eval))
				else:
					yield (index, self.function(*self.function_args, params=self.__index_to_dict(index, ranges_eval), **self.function_kwargs))
				if self.progress is not False:
					if self.progress is True:
						self.__print_progress_fallback(len(indices), i + 1, start_time)
					else:
						self.progress(len(indices), i + 1, start_time)

	def __print_progress_fallback(self, total, completed, start_time):
		progress = float(completed) / total

		sys.stderr.write("\r %3d%% | %d of %d | Memory usage: %.2f MB" % (
								progress * 100,
								completed,
								total,
								resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024.)
						)

		if progress > 0:
			delta = datetime.datetime.now() - start_time
			delta = datetime.timedelta(delta.total_seconds() / 24 / 3600 * (1 - progress) / progress)
			sys.stderr.write(" | Remaining: %02dd:%02dh:%02dm:%02ds" % (
					delta.days,
					delta.seconds / 3600,
					delta.seconds / 60 % 60,
					delta.seconds % 60
				)
			)

		if total == completed:
			sys.stderr.write('\n')

		sys.stderr.flush()
